from numpy.lib.shape_base import take_along_axis
from suzieq.poller.services.service import Service
from suzieq.utils import convert_macaddr_format_to_colon, expand_ios_ifname
import re
import numpy as np


class LldpService(Service):
    """LLDP service. Different class because of munging ifname"""

    def _common_data_cleaner(self, processed_data, raw_data):

        drop_indices = []

        for i, entry in enumerate(processed_data):

            if not entry:
                continue

            if not entry.get('ifname', ''):
                drop_indices.append(i)
                continue

            self._common_cleaner(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _common_cleaner(self, entry):
        chassis_type = entry.get('_chassisType', '')
        if not isinstance(chassis_type, str):
            if chassis_type == 7:
                chassis_type = 'peerHostname'
            else:
                chassis_type = 'unknown'
        chassis_type = chassis_type.lower()
        if chassis_type == 'mac address':
            entry['peerHostname'] = convert_macaddr_format_to_colon(
                entry.get('peerHostname', '0000.0000.0000'))

        subtype = str(entry.get('subtype', '')).lower()
        if subtype in ["interface name", '', '5', '7']:
            entry['peerMacaddr'] = '00:00:00:00:00:00'
            entry['peerIfindex'] = 0
            entry['subtype'] = 'interface name'  # IOS* don't provide subtype
        if subtype == 'mac address':
            entry['peerMacaddr'] = convert_macaddr_format_to_colon(
                entry.get('peerIfname', '0000.0000.0000'))
            entry['peerIfname'] = '-'
            entry['peerIfindex'] = 0
            entry['subtype'] = 'mac address'
        elif subtype.startswith(('locally', '1')):
            entry['peerIfindex'] = entry['peerIfname']
            entry['peerIfname'] = '-'
            entry['peerMacaddr'] = '00:00:00:00:00:00'
            entry['subtype'] = 'locally assigned'

    def _clean_nxos_data(self, processed_data, raw_data):

        drop_indices = []
        entries = {}

        for i, entry in enumerate(processed_data):
            entry['peerHostname'] = re.sub(r'\(.*\)', '',
                                           entry['peerHostname'])
            entry['ifname'] = re.sub(
                r'^Eth?(\d)', 'Ethernet\g<1>', entry['ifname'])

            if entry['ifname'] in entries:
                # Description is sometimes filled in with CDP, but not LLDP
                if not entry.get('description', ''):
                    old_entry = processed_data[entries[entry['ifname']]]
                    entry['description'] = old_entry.get('description', '')
                    entry['subtype'] = old_entry.get('subtype', '')
                    drop_indices.append(entries[entry['ifname']])
            else:
                entries[entry['ifname']] = i

            if entry.get('protocol', '') == 'cdp':
                entry['subtype'] = 'interface name'

            entry['peerIfname'] = expand_ios_ifname(
                entry.get('peerIfname', ''))

            if entry.get('mgmtIP', '') == "not advertised":
                entry['mgmtIP'] = ''  # make it consistent with other NOS
            self._common_cleaner(entry)
            if not entry['peerHostname']:
                drop_indices.append(i)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_junos_data(self, processed_data, raw_data):

        drop_indices = []
        for i, entry in enumerate(processed_data):
            if not entry.get('ifname', ''):
                drop_indices.append(i)
                continue

            self._common_cleaner(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_eos_data(self, processed_data, raw_data):

        drop_indices = []

        for i, entry in enumerate(processed_data):
            if not entry['ifname']:
                drop_indices.append(i)

            subtype = entry.get('subtype', '')
            if subtype == 'interfaceName':
                entry['subtype'] = 'interface name'
            elif subtype == 'macAddress':
                entry['subtype'] = 'mac address'
            elif subtype == "local":
                entry['subtype'] = 'locally assigned'

            self._common_cleaner(entry)
            if not entry['peerHostname']:
                drop_indices.append(i)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_cumulus_data(self, processed_data, raw_data):

        drop_indices = []

        for i, entry in enumerate(processed_data):
            if not entry['ifname']:
                drop_indices.append(i)
            subtype = entry.get('subtype', '')
            if subtype == "ifname":
                entry['subtype'] = "interface name"
            elif subtype == "local":
                entry['subtype'] = "locally assigned"

            self._common_cleaner(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_linux_data(self, processed_data, raw_data):

        drop_indices = []
        for i, entry in enumerate(processed_data):
            if not entry['ifname']:
                drop_indices.append(i)
            subtype = entry.get('subtype', '')
            if subtype == 'ifname':
                entry['subtype'] = 'interface name'
            elif subtype == 'local':
                entry['subtype'] = 'locally assigned'
            elif subtype == 'mac':
                entry['subtype'] = 'mac address'
            self._common_cleaner(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_iosxr_data(self, processed_data, raw_data):

        drop_indices = []
        for i, entry in enumerate(processed_data):
            if not entry['ifname']:
                drop_indices.append(i)
            self._common_cleaner(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_iosxe_data(self, processed_data, raw_data):

        drop_indices = []
        for i, entry in enumerate(processed_data):

            if not entry['ifname']:
                drop_indices.append(i)
            for field in ['ifname', 'peerIfname']:
                entry[field] = expand_ios_ifname(entry[field])
                if ' ' in entry.get(field, ''):
                    entry[field] = entry[field].replace(' ', '')
            self._common_cleaner(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_ios_data(self, processed_data, raw_data):
        return self._clean_iosxe_data(processed_data, raw_data)

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)



from .engineobj import SqPandasEngine


class MlagObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'mlag'

    def summarize(self, **kwargs):
        """Summarize MLAG info"""

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('uniqueSystemIdCnt', 'systemId', 'nunique')
        ]

        self._summarize_on_add_with_query = [
            ('devicesWithfailedStateCnt', 'state != "active"', 'state'),
            ('devicesWithBackupInactiveCnt', 'state == "active"', 'backupActive')
        ]

        self._summarize_on_add_stat = [
            ('mlagNumDualPortsStat', 'state == "active"', 'mlagDualPortsCnt'),
            ('mlagNumSinglePortStat', 'state == "active"', 'mlagSinglePortsCnt'),
            ('mlagNumErrorPortStat', 'state == "active"', 'mlagErrorPortsCnt')
        ]

        return super().summarize(**kwargs)



# List of Exceptions specific to Suzieq, across all the modules


class NoLLdpError(Exception):
    pass


class EmptyDataframeError(Exception):
    pass


class PathLoopError(Exception):
    pass


class DBReadError(Exception):
    pass


class DBNotFoundError(Exception):
    pass


class UserQueryError(Exception):
    pass


class UnknownDevtypeError(Exception):
    pass



from .engineobj import SqPandasEngine


class ArpndObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'arpnd'

    def summarize(self, **kwargs):
        """Summarize ARPND info across namespace"""
        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('arpNdEntriesCnt', 'ipAddress', 'count'),
            ('macaddrCnt', 'macaddr', 'count'),
            ('oifCnt', 'oif', 'count'),
            ('uniqueOifCnt', 'oif', 'nunique')]

        self._summarize_on_add_with_query = [
            ('remoteEntriesCnt', 'state == "remote"', 'ipAddress'),
            ('staticEntriesCnt', 'state == "permanent"', 'ipAddress'),
            ('failedEntryCnt', 'state == "failed"', 'ipAddress')]

        self._check_empty_col = 'arpNdEntriesCnt'
        return super().summarize(**kwargs)



#!/usr/bin/env python

import sys
import pyarrow.parquet as pq
import pyarrow as pa
from suzieq.utils import Schema, SchemaForTable
import logging
import argparse
import pandas as pd
from pathlib import Path


def convert_dir(input_dir: str, output_dir: str, svcschema: SchemaForTable):
    """Convert the data into a single file and write it out"""

    defaults = {
        pa.string(): "",
        pa.int32(): 0,
        pa.int64(): 0,
        pa.float32(): 0.0,
        pa.float64(): 0.0,
        pa.date64(): 0.0,
        pa.bool_(): False,
        pa.list_(pa.string()): ['-'],
        pa.list_(pa.int64()): [],
    }

    df = pd.read_parquet(input_dir, use_legacy_dataset=True)
    sqschema = svcschema.get_raw_schema()
    arrow_schema = svc_schema.get_arrow_schema()

    for column in filter(lambda x: x['name'] not in df.columns, sqschema):
        df[column['name']] = column.get('default', defaults[column['type']])

    # convert all dtypes to whatever is desired
    for column in df.columns:
        if column in arrow_schema:
            df[column] = df[column].astype(arrow_schema.field(column)
                                           .type.to_pandas_dtype())

    # If there's the original ifname saved up, then eliminate this unnecessary
    # field as this model is no longer necessary

    if 'origIfname' in df.columns:
        if 'ifname' in df.columns:
            df = df.drop(columns=['ifname']) \
                   .rename(columns={'origIfname': 'ifname'})
        elif 'oif' in df.columns:
            df = df.drop(columns=['oif']) \
                   .rename(columns={'origIfname': 'oif'})

    table = pa.Table.from_pandas(df, schema=arrow_schema,
                                 preserve_index=False)
    partition_cols = svcschema.get_partition_columns()

    if 'norifcnReason' in df.columns:
        df.rename({'notifcnReason': 'notificnReason'}, inplace=True)

    pq.write_to_dataset(
        table,
        root_path=output_dir,
        partition_cols=partition_cols,
        version="2.0",
        compression='ZSTD',
        row_group_size=100000,
    )

    logger.info(f'Wrote converted {input_dir}')


if __name__ == "__main__":
    if len(sys.argv) < 4:
        print('Usage: convert_parquet <input dir> <output_dir> <schema_dir>')
        sys.exit(1)

    input_dir = Path(sys.argv[1])
    output_dir = sys.argv[2]
    schemas = Schema(sys.argv[3])
    service = input_dir.parts[-1]
    svc_schema = SchemaForTable(service, schema=schemas)

    logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
    logger = logging.getLogger('sq-converter')
    convert_dir(input_dir, output_dir, svc_schema)



from pandas import DataFrame
from suzieq.sqobjects.basicobj import SqObject


class NetworkObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='network', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'version',
                                'os', 'model', 'vendor', 'columns', 'query_str']
        self._valid_find_args = ['namespace', 'hostname', 'vrf', 'vlan',
                                 'address', 'query_str']

    def find(self, **kwargs):

        addr = kwargs.get('address', '')
        asn = kwargs.get('asn', '')

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        if not addr and not asn:
            raise AttributeError('Must specify address or asn')

        try:
            self._check_input_for_valid_args(self._valid_find_args, **kwargs)
        except Exception as error:
            df = DataFrame({'error': [f'{error}']})
            return df

        return self.engine.find(**kwargs)



#!/usr/bin/env python3
"""
Generate AVRO schema files from the service definition and textfsm files
"""

import sys
import os
import re
import yaml
import json
import ast
from pathlib import Path


def get_field_set(svc_def, textfsm_dir):
    """Return the dict of fields with types given a service definition file"""

    field_set = set()
    defn_list = svc_def.get('apply', [])

    for entry in defn_list:
        nfn = defn_list[entry].get('normalize', None)
        if nfn:
            plist = re.split('''/(?=(?:[^'"]|'[^']*'|"[^"]*")*$)''', nfn)
            for elem in plist:
                try:
                    nfn_fld_list = ast.literal_eval(elem)
                    nfn_flds = [
                        re.match(r'\s*(?:add\(|div\(|mul\(|sub\()*(\w+)',
                                 fld.split(':')[1]).group(1)
                        for fld in nfn_fld_list if type(fld) is str
                        ]
                    field_set.update(nfn_flds)
                except (ValueError, SyntaxError):
                    words = elem.split(':')
                    if len(words) > 1:
                        field_set.add(re.match(r'\s*(\w+)',
                                               words[1]).group(1))
        else:
            nfn = defn_list[entry].get('textfsm', None)
            if nfn:
                with open(textfsm_dir + '/' + nfn, 'r') as f:
                    lines = f.readlines()
                for line in lines:
                    if line.startswith('Value'):
                        g = re.match(r'Value\s+(\w+)\s*(\w*)', line).groups()
                        if g[-1]:
                            fldname = g[-1]
                        else:
                            fldname = g[0]
                        field_set.add(fldname)

    return field_set


def write_avro_schema(name, fldset, keys, show_fields, type, output_dir):
    """Generate an AVRO schema file given a set of fields.
    We attempt to derive the type if we can. The logic followed for
    automatically deriving type is as follows:
    - if the field name ends in time/timestamp, type is float
    - if the field name ends in name, type is string
    - if the field name ends in list, type is array whose elements are string
    - if the field name ends in Cnt/Count, type is int
    """
    schema = {
        "namespace": "suzieq",
        "name": name,
        "type": "record",
        "fields": []
    }

    if type == 'counters':
        schema.update({'recordType': 'counters'})

    for fld in fldset:
        field = {'name': fld}

        if fld.endswith('Cnt') or fld.endswith('Count'):
            field['type'] = 'long'
        elif fld.endswith('List'):
            field['type'] = {'type': 'array',
                             'items': {'type': 'string',
                                       'name': fld.split('List')[0]}}
        elif (fld.endswith('Time') or fld.endswith('Timestamp') or
              fld.endswith('time')):
            field['type'] = 'timestamp'
        elif fld.endswith('name') or fld.endswith('Name') or fld == 'user':
            field['type'] = 'string'
        elif type == 'counters':
            field['type'] = 'long'
        else:
            field['type'] = 'string'

        if fld in keys:
            field['key'] = keys.index(fld) + 2

        if fld in show_fields:
            field['display'] = show_fields.index(fld) + 2

        schema['fields'].append(field)
        field = {}

    # Add the default fields we add to every record
    schema['fields'].append({'name': 'hostname', 'type': 'string',
                             'key': 1, 'display': 1})
    schema['fields'].append({'name': 'namespace', 'type': 'string',
                             'key': 0, 'display': 0})
    schema['fields'].append({'name': 'timestamp', 'type': 'timestamp'})
    schema['fields'].append({'name': 'active', 'type': 'boolean'})

    filename = output_dir + '/' + name + '.avsc'
    # We do require users to cleanup this auto-generated schema.
    # Don't make them do the work multiple times, save the original
    # type if defined, or default if not defined in the new schema
    if os.path.isfile(filename):
        with open(filename, 'r') as f:
            old_schema = json.loads(f.read())

        oldflds = {v['name']: i
                   for i, v in enumerate(old_schema['fields'])}
        newflds = {v['name']: i
                   for i, v in enumerate(schema['fields'])}

        for fld in oldflds.keys():
            if fld in newflds:
                schema['fields'][newflds[fld]]['type'] = old_schema['fields'][oldflds[fld]]['type']
                if (old_schema['fields'][oldflds[fld]].get('default', None)
                    and not schema['fields'][newflds[fld]].get('default',
                                                               None)):
                    schema['fields'][newflds[fld]]['default'] = old_schema['fields'][oldflds[fld]]['default']

    with open(filename, 'w') as f:
        f.write(json.dumps(schema, indent=4))

    return


if __name__ == '__main__':

    if len(sys.argv) < 2:
        print('Usage: gen_schema <service file|service dir> <directory '
              'to write schema to> [<config directory>]')
        sys.exit(1)

    if not os.path.isdir(sys.argv[2]):
        print('ERROR: {} must be a directory. Aborting', sys.argv[2])
        sys.exit(1)

    textfsm_dir = None

    p = Path(sys.argv[1])
    if p.is_file():
        files = [p]
    elif p.is_dir():
        files = p.glob('*.yml')
        textfsm_dir = p.name
    else:
        files = []

    if len(sys.argv) > 3:
        textfsm_dir = sys.argv[3]

    if not textfsm_dir:
        print('ERROR: Specify parent directory of textfsm template dir')
        sys.exit(1)

    if not Path(textfsm_dir).is_dir():
        print('ERROR: Specify parent directory of textfsm template dir')
        sys.exit(1)

    for file in files:
        with file.open() as f:
            data = yaml.safe_load(f.read())
            fldset = get_field_set(data, textfsm_dir)
            write_avro_schema(data['service'], fldset, data.get('keys', []),
                              data.get('show-fields', []),
                              data.get('type', None), sys.argv[2])



import re
from dateparser import parse
from datetime import datetime

from suzieq.poller.services.service import Service
from suzieq.utils import (expand_nxos_ifname, get_timestamp_from_cisco_time,
                          get_timestamp_from_junos_time)


import numpy as np


class RoutesService(Service):
    """routes service. Different class because vrf default needs to be added"""

    def clean_json_input(self, data):
        """Junos JSON data for some older ver needs some work"""

        devtype = data.get("devtype", None)
        if devtype.startswith('junos'):
            data['data'] = data['data'].replace('}, \n    }\n', '} \n    }\n')
            return data['data']

    def _fix_ipvers(self, entry):
        if ':' in entry['prefix']:
            entry['ipvers'] = 6
        else:
            entry['ipvers'] = 4

    def _common_data_cleaner(self, processed_data, raw_data):
        for entry in processed_data:
            self._fix_ipvers(entry)

        return processed_data

    def _clean_eos_data(self, processed_data, raw_data):
        '''Massage EVPN routes'''
        for entry in processed_data:
            if entry['nexthopIps']:
                nexthop = entry['nexthopIps'][0]
                if 'vtepAddr' in nexthop:
                    nexthop = entry['nexthopIps'][0]
                    entry['nexthopIps'] = [nexthop['vtepAddr']]
                    entry['oifs'] = ['_nexthopVrf:default']
            elif entry.get('_vtepAddr', []):
                entry['nexthopIps'] = entry['_vtepAddr']
                entry['oifs'] = len(entry['nexthopIps']) * \
                    ['_nexthopVrf:default']
            entry['protocol'] = entry['protocol'].lower()
            entry['preference'] = int(entry.get('preference', 0))
            entry['metric'] = int(entry.get('metric', 0))
            self._fix_ipvers(entry)

        return processed_data

    def _clean_linux_data(self, processed_data, raw_data):
        """Clean Linux ip route data"""

        for entry in processed_data:
            entry["vrf"] = entry["vrf"] or "default"
            entry["metric"] = entry["metric"] or 20
            entry['preference'] = entry['metric']
            for ele in ["nexthopIps", "oifs"]:
                entry[ele] = entry[ele] or [""]
            entry['hardwareProgrammed'] = 'unknown'
            entry["weights"] = entry["weights"] or [1]
            if entry['prefix'] == 'default':
                if any(':' in x for x in entry['nexthopIps']):
                    entry['prefix'] = '::0/0'
                else:
                    entry['prefix'] = '0.0.0.0/0'

            self._fix_ipvers(entry)

            if '/' not in entry['prefix']:
                if entry['ipvers'] == 6:
                    entry['prefix'] += '/128'
                else:
                    entry['prefix'] += '/32'
            if not entry["action"]:
                entry["action"] = "forward"
            elif entry["action"] == "blackhole":
                entry["oifs"] = ["blackhole"]

            entry['inHardware'] = True  # Till the offload flag is here

        return processed_data

    def _clean_cumulus_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)

    def _clean_junos_data(self, processed_data, raw_data):
        """Clean VRF name in JUNOS data"""

        drop_entries_idx = []
        prefix_entries = {}

        for i, entry in enumerate(processed_data):

            if '_vtepAddr' in entry:
                # Get the entry from the prefix DB. We have the correct NH
                pentry = prefix_entries.get(entry['prefix'], None)
                if not pentry:
                    drop_entries_idx.append(i)
                    continue
                pentry['nexthopIps'] = [entry['_vtepAddr']]
                pentry['oifs'] = ['_nexthopVrf:default']
                drop_entries_idx.append(i)
                continue

            vrf = entry.pop("vrf")[0]['data']
            if vrf == "inet.0":
                vrf = "default"
                vers = 4
            elif vrf == "inet6.0":
                vrf = "default"
                vers = 6
            else:
                words = vrf.split('.')
                vrf = words[0]
                family = words[1]
                if family == "inet":
                    vers = 4
                elif family == "inet6":
                    vers = 6
            entry['vrf'] = vrf
            entry['ipvers'] = vers

            if entry['_localif']:
                entry['oifs'] = [entry['_localif']]

            entry['protocol'] = entry['protocol'].lower()
            if entry['_rtlen'] != 0:
                drop_entries_idx.append(i)

            if entry.get('asPathList', []):
                entry['asPathList'] = entry['asPathList'].split()

            prefix_entries[entry['prefix']] = entry

            lastChange = entry.get('statusChangeTimestamp', [''])
            if lastChange:
                entry['statusChangeTimestamp'] = get_timestamp_from_junos_time(
                    lastChange, raw_data[0]['timestamp']/1000)
            else:
                entry['statusChangeTimestamp'] = 0

            entry['active'] = entry['_activeTag'] in ['*', '@', '#']

            entry['preference'] = int(entry.get('preference', 0))
            entry['metric'] = int(entry.get('metric', 0))
            if entry.get('nexthopIps', '') == [None]:
                entry['nexthopIps'] = ['']

            entry['action'] = entry.get('action', '').lower()
            if entry['action'] == "discard":
                entry['action'] = 'drop'
            entry.pop('_localif')
            entry.pop('_activeTag')

        processed_data = np.delete(processed_data,
                                   drop_entries_idx).tolist()

        return processed_data

    def _clean_nxos_data(self, processed_data, raw_data):

        drop_indices = []

        for i, entry in enumerate(processed_data):
            if 'prefix' not in entry or not entry['prefix']:
                drop_indices.append(i)
                continue

            entry['protocol'] = entry.get('protocol', '').split('-')[0]

            entry['metric'] = [int(x) if x is not None else 0
                               for x in entry.get('metric', [])]
            entry['metric'] = entry['metric'][0]
            try:
                entry['preference'] = int(entry.get('preference', [0])[0])
            except ValueError:
                entry['preference'] = 0

            oiflist = []
            for oif in entry['oifs']:
                if oif:
                    oif = expand_nxos_ifname(oif)
                    oiflist.append(oif)
            entry['oifs'] = oiflist

            nhips = []
            nhoifs = []
            for ent in entry['nexthopIps']:
                if '%' in ent:
                    nhip, vrf = ent.split('%')
                    nhips.append(nhip)
                    nhoifs.append(f'_nexthopvrf:{vrf}')

            if nhips:
                entry['nexthopIps'] = nhips
                entry['oifs'] = nhoifs

            if not entry['oifs']:
                oiflist = []
                for nhv in entry.get('_nexthopVrf', []):
                    if nhv:
                        oiflist.append(f'_nexthopVrf:{nhv}')
                entry['oifs'] = oiflist

            if 'discard' in entry.get('_routeAction', []):
                entry['action'] = 'drop'
            else:
                entry['action'] = 'forward'

            # Right now we only store the timestamp of the first NH change
            # The correct thing to do is to take all NH timestamps, and take
            # the latest one: TODO
            lastChange = entry.get('statusChangeTimestamp', [''])[0]
            if lastChange:
                entry['statusChangeTimestamp'] = get_timestamp_from_cisco_time(
                    lastChange, raw_data[0]['timestamp']/1000)
            else:
                entry['statusChangeTimestamp'] = 0

            self._fix_ipvers(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_iosxr_data(self, processed_data, raw_data):

        proto_map = {'C': ['connected', ''], 'S': ['static', ''],
                     'R': ['rip', ''], 'B': ['bgp', ''],
                     'D': ['eigrp', ''], 'EX': ['eigrp', 'external'],
                     'O': ['ospf' ''], 'O IA': ['ospf', 'interarea'],
                     'O N1': ['ospf', 'nssa-ext1'],
                     'O N2': ['ospf', 'nssa-ext2'],
                     'O E1': ['ospf', 'external-1'],
                     'O E2': ['ospf', 'external-2'], 'E': ['igp', ''],
                     'i': ['isis', ''], 'i L1': ['isis', 'level-1'],
                     'i L2': ['isis', 'level-2'],
                     'i ia': ['isis', 'interarea'],
                     'i su': ['isis', 'summary-null'], 'U': ['static', 'user'],
                     'o': ['odr', ''], 'L': ['local', ''], 'G': ['dagr', ''],
                     'l': ['lisp', ''], 'M': ['mobile', ''],
                     'a': ['application', ''], 'r': ['rpl', ''],
                     't': ['te', ''],
                     'A': ['access', ''], '!': ['backup', '']}

        for entry in processed_data:
            entry['protocol'] = proto_map.get(entry.get('protocol', '')[0],
                                              entry.get('protocol', ''))[0]
            if not entry['vrf']:
                entry['vrf'] = 'default'
            if entry.get('nexthopvrf', ''):
                entry['oifs'] = [f'_nexthopVrf:{entry["nexthopvrf"]}']
            if entry.get('metric', '') == '':
                entry['metric'] = 0
            if entry.get('preference', '') == '':
                entry['preference'] = 0
            self._fix_ipvers(entry)
            entry['hardwareProgrammed'] = 'unknown'
            if not entry.get('action', ''):
                entry['action'] = 'forward'
            else:
                entry['action'] = entry.get('action', '').lower()

            lastchange = entry.get('statusChangeTimestamp', '')
            if lastchange:
                if re.match(r'^\d{2}:\d{2}:\d{2}$', lastchange):
                    lastchange = lastchange.split(':')
                    lastchange = (f'{lastchange[0]} hour '
                                  f'{lastchange[1]}:{lastchange[2]} mins ago')
                lastchange = parse(
                    lastchange,
                    settings={'RELATIVE_BASE':
                              datetime.fromtimestamp(
                                  (raw_data[0]['timestamp'])/1000), })
            if lastchange:
                entry['statusChangeTimestamp'] = lastchange.timestamp()*1000
            else:
                entry['statusChangeTimestamp'] = 0

        return processed_data

    def _clean_iosxe_data(self, processed_data, raw_data):
        processed_data = self._clean_iosxr_data(processed_data, raw_data)

        # Some more IOSXE fixes including:
        #  * lowercasing IPv6 addresses
        #  * adding / to host prefixes
        #  * adding IPv4 netmask if not included in the prefix
        for entry in processed_data:
            if ':' in entry['prefix']:
                entry['prefix'] = entry['prefix'].lower()
                if '/' not in entry['prefix']:
                    entry['prefix'] += '/128'
            elif '/' not in entry['prefix']:
                if 'netmask' in entry:
                    entry['prefix'] += '/' + entry['netmask']
                    del entry['netmask']
                else:
                    entry['prefix'] += '/32'
            newnexthops = []
            for ele in entry['nexthopIps']:
                if ':' in ele:
                    newnexthops.append(ele.lower())
                else:
                    newnexthops.append(ele)
            entry['nexthopIps'] = newnexthops

        return processed_data

    def _clean_ios_data(self, processed_data, raw_data):
        return self._clean_iosxe_data(processed_data, raw_data)



#!/usr/bin/env python3

import sys
import shlex


def process_ansible_inventory(filename, namespace='default'):
    """Read an Ansible inventory file and produce device list for suzieq

    :param filename: Ansible inventory filename
    :type filename: string

    :param namespace: The namespace associated with the data from these devices
    :type namespace: string

    :rtype: list:
    :return: List of devices as extracted from the Ansible file
    """

    with open(filename, 'r') as f:
        lines = f.readlines()

    hostsdata = []
    for line in lines:
        host = None
        hostline = ''
        out = shlex.split(line)
        if 'ansible_network_os=eos' in out:
            transport = 'https'
            addnl_info = 'devtype=eos'
        else:
            transport = 'ssh'
            addnl_info = ''
        for elem in out:
            if elem.startswith('ansible_host='):
                k, v = elem.split('=')

                if v == '127.0.0.1':
                    host = v
                    port = 0
                    continue
                else:
                    hostline = ('    - url: {}://vagrant@{} {}'
                                .format(transport, v, addnl_info))
            if elem.startswith('ansible_port='):
                k, v = elem.split('=')
                if host:
                    hostline = (
                        '    - url: ssh://vagrant@{}:{}'.format(host, v))

            if elem.startswith('ansible_ssh_private_key_file') and transport == 'ssh':
                k, v = elem.split('=')
                if hostline:
                    hostline += ' keyfile={}'.format(v.strip())
                else:
                    addnl_info += ' keyfile={}'.format(v.strip())

        hostsdata.append(hostline)

    hostsdata.insert(0, '- namespace: {}'.format(namespace))
    hostsdata.insert(1, '  hosts:')

    return hostsdata


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print('Usage: genhosts <Ansible inventory file> <output file> <DC name>')
        sys.exit(1)

    hostsdata = process_ansible_inventory(sys.argv[1], sys.argv[3])

    out = '\n'.join(hostsdata)

    with open(sys.argv[2], 'w') as f:
        f.write(out)



import os
from importlib.util import find_spec
from importlib import import_module
from pathlib import Path
import inspect

name = "sqengines"


def get_sqengine(engine_name: str, table_name: str):

    # Lets load the available engines
    try:
        eng_mod = import_module(f'suzieq.engines.{engine_name}')
    except ModuleNotFoundError:
        return None

    for mbr in inspect.getmembers(eng_mod):
        if mbr[0] == 'get_engine_object' and inspect.isfunction(mbr[1]):
            return mbr[1]

    return None


__all__ = [get_sqengine]



from .engineobj import SqPandasEngine
from dataclasses import dataclass
import os

import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from suzieq.exceptions import EmptyDataframeError


# TODO:
#  topology for different VRFs?
#  iBGP vs eBGP?
#  color by device type?
#  physical topology without LLDP -- is this possible?
#  how to draw multiple topologies
#  be able to ask if a node has neighbors by type (physical, overlay, protocol, etc)
#  questions
#    * without knowing hierarchy, labels or tags it's unclear how to group things for good picture
# how could we add state of connection (like Established) per protocol

graph_output_dir = '/tmp/suzieq-graphs'


class TopologyObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'topology'

    def _init_dfs(self, namespaces):
        """Initialize the dataframes used"""

        self._if_df = self._get_table_sqobj('interfaces') \
            .get(namespace=namespaces, state="up",
                 columns=['namespace', 'hostname', 'ifname', 'ipAddressList',
                          'ip6AddressList', 'state', 'type', 'master', 'macaddr'])

        if self._if_df.empty:
            raise EmptyDataframeError(f"No interface found for {namespaces}")
        self._if_df['vrf'] = self._if_df.apply(
            lambda x: x['master'] if x['type'] not in ['bridge', 'bond_slave']
            else 'default', axis=1)
        self._if_df['vrf'] = np.where(self._if_df['vrf'] == '', 'default',
                                      self._if_df['vrf'])
        self._if_df = self._if_df.explode('ipAddressList') \
                                 .explode('ip6AddressList') \
                                 .fillna({'ipAddressList': '0.0.0.0/0',
                                          'ip6AddressList': '::0/0'})
        self._if_df['plen'] = self._if_df.ipAddressList.str.split('/').str[1]
        self._if_df['ipAddress'] = self._if_df.ipAddressList.str.split(
            '/').str[0]
        self._if_df = self._if_df.drop(
            columns=['ipAddressList', 'ip6AddressList'])

        self.nses = self._if_df['namespace'].unique()

    def get(self, **kwargs):
        self._namespaces = kwargs.get("namespace", self.ctxt.namespace)
        hostname = kwargs.pop('hostname', [])
        user_query = kwargs.pop('query_str', '')
        polled = kwargs.pop('polled', '')
        via = kwargs.pop('via', [])
        ifname = kwargs.pop('ifname', '')
        peerHostname = kwargs.pop('peerHostname', [])
        columns = kwargs.pop('columns', ['default'])

        self._init_dfs(self._namespaces)
        self.lsdb = pd.DataFrame()
        self._a_df = pd.DataFrame()
        self._ip_table = pd.DataFrame()

        self.services = [
            Services('lldp', {}, ['ifname', 'vrf'],
                     self._augment_lldp_show),
            Services('arpnd', {},
                     ['ipAddress', 'macaddr', 'ifname', 'vrf', 'arpndBidir'],
                     self._augment_arpnd_show),
            Services('bgp', {'state': 'Established',
                             'columns': ['*']},
                     ['vrf'],
                     self._augment_bgp_show),
            # Services('evpnVni', {}, 'peerHostname',
            #    'vni', self._augment_evpnvni_show),
            Services('ospf', {'state': 'full'},
                     ['ifname', 'vrf'], self._augment_ospf_show),
        ]

        if via:
            self.services = [x for x in self.services if x.name in via]
        key = 'peerHostname'
        for srv in self.services:
            if 'columns' not in srv.extra_args:
                srv.extra_args['columns'] = columns
            df = self._get_table_sqobj(srv.name).get(
                **kwargs,
                **srv.extra_args
            )

            if not df.empty:
                if srv.augment:
                    df = srv.augment(df)
                cols = ['namespace', 'hostname', key]
                if srv.extra_cols:
                    cols = cols + srv.extra_cols
                df = df[cols]
                df.insert(len(df.columns), srv.name, True)
                if self.lsdb.empty:
                    self.lsdb = df
                else:
                    self.lsdb = self.lsdb.merge(df, how='outer')

        self._find_polled_neighbors(polled)
        if self.lsdb.empty:
            return self.lsdb

        if 'ipAddress' not in self.lsdb.columns:
            self.lsdb['ipAddress'] = ''

        self.lsdb = self.lsdb[~self.lsdb.hostname.isna()] \
                        .drop_duplicates() \
                        .rename({'ipAddress': 'peerIP', 'macaddr': 'peerMac'},
                                axis=1, errors='ignore') \
                        .dropna(subset=['peerHostname', 'peerIP'], how='all') \
                        .fillna({'peerHostname': 'unknown',
                                 'ifname': 'unknown', 'arpnd': False,
                                 'peerIP': '', 'peerMac': '',
                                 'arpndBidir': False,
                                 'bgp': False, 'ospf': False,
                                 'lldp': False, 'vrf': 'N/A'})

        self.lsdb['vrf'] = np.where(self.lsdb.vrf == "bridge", "-",
                                    self.lsdb.vrf)
        self.lsdb['peerHostname'] = np.where(
            self.lsdb.peerHostname == "unknown", self.lsdb.peerIP,
            self.lsdb.peerHostname)
        self.lsdb = self.lsdb.drop(columns=['peerIP', 'peerMac'],
                                   errors='ignore')
        self.lsdb = self.lsdb.query('peerHostname != ""')

        # Apply the appropriate filters
        if not self.lsdb.empty:
            if hostname:
                self.lsdb = self.lsdb.query(f'hostname.isin({hostname})')
            if peerHostname:
                self.lsdb = self.lsdb.query(
                    f'peerHostname.isin({peerHostname})')
            if ifname:
                self.lsdb = self.lsdb.query(f'ifname.isin({ifname})')

        if user_query and not self.lsdb.empty:
            self.lsdb = self.lsdb.query(user_query)

        cols = self.lsdb.columns.tolist()
        if 'timestamp' in self.lsdb.columns:
            cols.remove('timestamp')
            cols.append('timestamp')

        return self.lsdb[cols].reset_index(drop=True)

    def _find_polled_neighbors(self, polled):
        if self.lsdb.empty:
            return

        devices = self._get_table_sqobj('device').get(namespace=self._namespaces,
                                                      columns=['namespace', 'hostname'])
        self.lsdb = devices.merge(self.lsdb, how='outer',
                                  indicator=True)
        self.lsdb = self.lsdb.rename(columns={'_merge': 'polled'})
        self.lsdb.polled = self.lsdb.polled == 'both'

        if polled and polled.lower() == 'true':
            self.lsdb = self.lsdb.query('polled == True')
        elif polled.lower() == 'false':
            self.lsdb = self.lsdb.query('polled == False')
        if self.lsdb.empty:
            self.lsdb = pd.DataFrame()

    def _create_graphs_from_lsdb(self):
        self.graphs = {}
        for ns, df in self.lsdb.groupby(by=['namespace']):
            attrs = [srv.name for srv in self.services if srv.name in df.columns]
            self.graphs[ns] = nx.from_pandas_edgelist(
                df, 'hostname', 'peerHostname', attrs, nx.MultiGraph)

    # TODO: eventually this needs to move to ospf after we figure out the
    #   schema augmentation story
    def _augment_ospf_show(self, df):
        if not df.empty:
            df = df.query('adjState != "passive"').reset_index(drop=True)

        return df

    def _augment_lldp_show(self, df):
        if not df.empty:
            df = df[df.peerHostname != '']
            df = df.merge(
                self._if_df[['namespace', 'hostname', 'ifname', 'vrf']],
                on=['namespace', 'hostname', 'ifname'])
        return df

    def _augment_bgp_show(self, df):
        return df

    def _augment_evpnvni_show(self, df):
        if not df.empty:
            df = df.explode('remoteVtepList').dropna(how='any')
            df = df.rename(columns={'remoteVtepList': 'peerIP'})
            df = df.merge(self.ip_table, on=['namespace', 'peerIP'],
                          how='left').dropna(how='any').drop_duplicates()
        return df

    def _augment_arpnd_show(self, df):

        if not df.empty:
            # weed out entries that are not reachable
            df = df.query('state != "failed"').reset_index(drop=True)
            # Get the VRF
            df = df.merge(
                self._if_df[['namespace', 'hostname', 'ifname', 'master']],
                left_on=['namespace', 'hostname', 'oif'],
                right_on=['namespace', 'hostname', 'ifname'],
                suffixes=['', '_y'], how='outer') \
                .rename(columns={'master': 'vrf'}) \
                .dropna(subset=['ipAddress'])

            df['vrf'] = np.where(df['vrf'] == '', 'default', df['vrf'])
            df = df.drop(columns=['vlan', 'oif', 'mackey', 'remoteVtepIp',
                                  'timestamp_y'], errors='ignore')

            # address are where I find the neighbor, these own the mac
            addr = self.address_df[['namespace', 'hostname', 'macaddr']] \
                .drop_duplicates() \
                .rename(columns={'hostname': 'peerHostname'})
            df = df.merge(addr, on=['namespace', 'macaddr'], how='left') \
                   .dropna(subset=['hostname', 'ipAddress']) \
                   .rename(columns={'oif': 'ifname'}) \
                   .drop_duplicates()

            # Use MAC table entries to find the local port for a MAC on an SVI
            mac_df = self._get_table_sqobj('macs')\
                         .get(namespace=self._namespaces, localOnly=True,
                              columns=['namespace', 'hostname', 'vlan',
                                       'macaddr', 'oif'])
            df = df.merge(mac_df,
                          on=['namespace', 'hostname', 'macaddr'],
                          how='outer') \
                .dropna(subset=['hostname'])
            df['ifname'] = np.where(
                df['oif'].isnull(), df['ifname'], df['oif'])

            df['arpndBidir'] = df.apply(
                lambda x, y: True
                if not y.query(f'namespace=="{x.namespace}" and '
                               f'hostname=="{x.peerHostname}" and '
                               f'peerHostname=="{x.hostname}"').empty
                else False, args=(df,), axis=1)

        self._arpnd_df = df
        return self._arpnd_df

    @property
    def address_df(self):
        if self._a_df.empty:
            self._a_df = self._if_df
        return self._a_df

    @property
    def ip_table(self):
        if self._ip_table.empty:
            addr = self.address_df
            if not addr.empty:
                self._ip_table = addr[['namespace',
                                       'hostname', 'ipAddressList']]
                self._ip_table = self._ip_table.explode(
                    'ipAddressList').dropna(how='any')
                self._ip_table = self._ip_table \
                                     .rename(
                                         columns={'ipAddressList': 'peerIP',
                                                  'hostname': 'peerHostname'})
                self._ip_table['peerIP'] = self._ip_table['peerIP'] \
                                               .str.replace("/.+", "")
                self._ip_table = self._ip_table[self._ip_table['peerIP'] != '-']
        return self._ip_table

    def summarize(self, **kwargs):
        self.get(**kwargs)
        if self.lsdb.empty:
            return self.lsdb

        self.ns = {}
        for i in self.nses:
            self.ns[i] = {}
        self._create_graphs_from_lsdb()
        self._analyze_lsdb_graph()

        return pd.DataFrame(self.ns)

    def _analyze_lsdb_graph(self):
        for ns in self.nses:
            for name in [srv.name for srv in self.services if
                         nx.get_edge_attributes(self.graphs[ns], srv.name)]:

                G = nx.Graph([(s, d, data) for s, d, data in
                              self.graphs[ns].edges(data=True) if data[name] == True])
                if G.nodes:
                    self.ns[ns][f'{name}_number_of_nodes'] = len(G.nodes)
                    self.ns[ns][f'{name}_number_of_edges'] = len(G.edges)
                    if not nx.is_connected(G):
                        self.ns[ns][f'{name}_is_fully_connected'] = False
                        self.ns[ns][f'{name}_center'] = False
                    else:
                        self.ns[ns][f'{name}_is_fully_connected'] = True
                        self.ns[ns][f'{name}_center'] = nx.barycenter(G)

                    self.ns[ns][f'{name}_self_loops'] = list(
                        nx.nodes_with_selfloops(G))

                    self.ns[ns][f'{name}_number_of_disjoint_sets'] = len(
                        list(nx.connected_components(G)))

                    self.ns[ns][f'{name}_degree_histogram'] = nx.degree_histogram(
                        G)

                    # if there are too many degrees than the column gets too big
                    if len(self.ns[ns][f'{name}_degree_histogram']) > 6:
                        self.ns[ns][f'{name}_degree_histogram'] = '...'

                else:
                    for k in [f'{name}_is_fully_connected', f'{name}_center',
                              f'{name}_self_loops', f'{name}_number_of_disjoint_sets',
                              f'{name}_degree_histogram', f'{name}_number_of_nodes',
                              f'{name}_number_of_edges']:
                        self.ns[ns][k] = None

    def _make_images(self):
        if not os.path.exists(graph_output_dir):
            os.makedirs(graph_output_dir)
        for ns in self.nses:
            pos = nx.spring_layout(self.graphs[ns])
            for name in [srv.name for srv in self.services if
                         nx.get_edge_attributes(self.graphs[ns], srv.name)]:
                edges = [(u, v) for u, v, d in self.graphs[ns].edges(data=True) if
                         d[name] == True]
                if len(edges) > 1:

                    nx.draw_networkx_nodes(
                        self.graphs[ns], pos=pos, node_size=25)
                    if len(self.graphs[ns].nodes) < 20:
                        nx.draw_networkx_labels(
                            self.graphs[ns], pos=pos, font_size=8)

                    nx.draw_networkx_edges(
                        self.graphs[ns], edgelist=edges, pos=pos)
                    plt.savefig(f"{graph_output_dir}/{ns}_{name}.png")
                    print(f"created {graph_output_dir}/{ns}_{name}.png")
                    plt.close()

    def unique(self, **kwargs):
        '''Unique values for topology'''
        columns = kwargs.pop('columns', ['default'])
        count = kwargs.pop("count", 0)
        query_str = kwargs.get('query_str', '')

        df = self.get(columns=['*'], **kwargs)
        column = columns[0]

        if not count:
            return (pd.DataFrame({f'{column}': df[column].unique()}))
        else:
            r = df[column].value_counts()
            return (pd.DataFrame({column: r})
                    .reset_index()
                    .rename(columns={column: 'numRows',
                                     'index': column})
                    .sort_values(column))


@dataclass(frozen=True)
class Services:
    name: str
    extra_args: dict
    extra_cols: list
    augment: any



from suzieq.sqobjects.basicobj import SqObject
import pandas as pd
from suzieq.utils import humanize_timestamp


class DeviceObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='device', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'os',
                                'vendor', 'model', 'status', 'version',
                                'query_str']

    def humanize_fields(self, df: pd.DataFrame, subset=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        # Convert the bootup timestamp into a time delta
        if 'bootupTimestamp' in df.columns:
            df['bootupTimestamp'] = humanize_timestamp(
                df['bootupTimestamp']*1000,
                self.cfg.get('analyzer', {}).get('timezone', None))
        return df



import time
from datetime import timedelta
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.ospf import OspfObj


@command("ospf", help="Act on OSPF data")
class OspfCmd(SqCommand):
    """OSPFv2 protocol information"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = ' ',
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=OspfObj,
        )

    @command("show")
    @argument(
        "ifname", description="Space separated list of interface names to qualify"
    )
    @argument("state", description="Select view based on status",
              choices=["full", "other", "passive"])
    @argument("vrf", description="Space separated list of VRFs to qualify")
    def show(self, ifname: str = "", state: str = "", vrf: str = ""):
        """Show OSPF interface and neighbor info
        """
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                vrf=vrf.split(),
                                ifname=ifname.split(),
                                columns=self.columns,
                                state=state,
                                query_str=self.query_str,
                                namespace=self.namespace,
                                )

        df = self.sqobj.humanize_fields(df)

        # Transform the lastChangeTime into human terms
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("summarize")
    @argument(
        "ifname", description="Space separated list of interface names to qualify"
    )
    @argument("vrf", description="Space separated list of VRFs to qualify")
    @argument("state", description="BGP neighbor state to qualify", choices=["full"])
    @argument(
        "type",
        description="Type of OSPF information to show",
        choices=["neighbor", "interface"],
    )
    @argument("groupby", description="Space separated list of fields to summarize on")
    def summarize(
        self,
        ifname: str = "",
        vrf: str = "",
        state: str = "",
        type: str = "neighbor",
        groupby: str = "",
    ):
        """
        Summarize OSPF data
        """
        self._init_summarize()
        return self._post_summarize()

    @command("assert")
    @argument("vrf", description="VRF to assert OSPF state in")
    @argument("status", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    def aver(self, vrf: str = "", status: str = 'all') -> pd.DataFrame:
        """
        Test OSPF runtime state is without errors
        """
        if self.hostname:
            df = pd.DataFrame(
                {'error': ['ERROR: You cannot specify hostname with assert']})
            return self._gen_output(df)
        now = time.time()
        df = self.sqobj.aver(
            vrf=vrf.split(),
            namespace=self.namespace,
            status=status,
        )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)



import time
import re
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.device import DeviceObj


@command("device", help="Act on device data")
class DeviceCmd(SqCommand):
    """Basic device information such as OS, version, model etc."""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",
            query_str: str = " ",
            columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=DeviceObj,
        )

    def _get(self, **kwargs):
        # Get the default display field names
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname, columns=self.columns,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                **kwargs,
                                )

        df = self.sqobj.humanize_fields(df)

        return df

    @command("show", help="Show device information")
    @argument("os", description="filter by NOS")
    @argument("vendor", description="filter by vendor")
    @argument("model", description="filter by model")
    @argument("version", description="filter by version")
    @argument("status", description="filter by polling status",
              choices=["dead", "alive",  "neverpoll"])
    def show(self, os: str = '', model: str = '', status: str = '',
             vendor: str = '', version: str = ''):
        """Show device info
        """
        if self.columns is None:
            return

        now = time.time()

        # Model has to be special cased because model names can have
        # spaces in them such as "Nexus9000 C9300v Chassis"
        if model:
            model = re.split(r"\s+(?=[^']*(?:'))", model)
        else:
            model = []
        df = self._get(os=os.split(), model=model, version=version,
                       vendor=vendor.split(), status=status.split())

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['suzieq',
 'suzieq.cli',
 'suzieq.cli.sqcmds',
 'suzieq.db',
 'suzieq.db.parquet',
 'suzieq.engines',
 'suzieq.engines.pandas',
 'suzieq.gui',
 'suzieq.gui.pages',
 'suzieq.poller',
 'suzieq.poller.nodes',
 'suzieq.poller.services',
 'suzieq.restServer',
 'suzieq.sqobjects',
 'suzieq.utilities']

package_data = \
{'': ['*'],
 'suzieq': ['config/*',
            'config/etc/*',
            'config/schema/*',
            'config/textfsm_templates/*'],
 'suzieq.gui': ['images/*'],
 'suzieq.gui.pages': ['help/*']}

install_requires = \
['PyYAML',
 'aiofiles',
 'aiohttp==3.7.4',
 'async-timeout',
 'asyncssh>=2.7,<3.0',
 'colorama>=0.4.4,<0.5.0',
 'dateparser>=1.0.0,<2.0.0',
 'faker>=4.1.1,<5.0.0',
 'fastapi>=0.65,<0.66',
 'graphviz>=0.15,<0.16',
 'jsonpath-ng>=1.5.1,<2.0.0',
 'matplotlib>=3.2.2,<4.0.0',
 'natsort>=7.1.1,<8.0.0',
 'netconan>=0.11.2,<0.12.0',
 'networkx>=2.4,<3.0',
 'pandas==1.2.5',
 'prompt-toolkit>2',
 'pyarrow==5.0.0',
 'python-nubia==0.2b5',
 'streamlit>=0.87.0,<0.88.0',
 'tabulate>=0.8.7,<0.9.0',
 'textfsm',
 'tzlocal<3.0',
 'uvicorn>=0.14.0,<0.15.0',
 'uvloop']

entry_points = \
{'console_scripts': ['sq-anonymizer = '
                     'suzieq.utilities.sq_anonymizer:anonymizer_main',
                     'sq-coalescer = '
                     'suzieq.utilities.sq_coalescer:coalescer_main',
                     'sq-poller = suzieq.poller.sq_poller:poller_main',
                     'sq-rest-server = '
                     'suzieq.restServer.sq_rest_server:rest_main',
                     'suzieq-cli = suzieq.cli.sq_cli:cli_main',
                     'suzieq-gui = suzieq.gui.sq_gui:gui_main']}

setup_kwargs = {
    'name': 'suzieq',
    'version': '0.15.5',
    'description': 'A framework and application for network observability',
    'long_description': "[![integration-tests](https://github.com/netenglabs/suzieq/workflows/integration-tests/badge.svg)](https://github.com/netenglabs/suzieq/actions/workflows/integration-tests.yml)\n[![GitHub release (latest by date)](https://img.shields.io/github/v/release/netenglabs/suzieq?logo=github&color=success)](https://github.com/netenglabs/suzieq/releases/latest)\n[![GitHub](https://img.shields.io/github/license/netenglabs/suzieq?logo=github&color=success)](LICENSE)\n[![Docker Image Version (latest by date)](https://img.shields.io/docker/v/netenglabs/suzieq?logo=docker&color=blue)](https://hub.docker.com/r/netenglabs/suzieq/tags?page=1&ordering=last_updated)\n[![Docker Image Size (latest by date)](https://img.shields.io/docker/image-size/netenglabs/suzieq?logo=docker&color=blue)](https://hub.docker.com/r/netenglabs/suzieq/tags?page=1&ordering=last_updated)\n[![Docker Pulls](https://img.shields.io/docker/pulls/netenglabs/suzieq?logo=docker&color=blue)](https://hub.docker.com/r/netenglabs/suzieq/tags?page=1&ordering=last_updated)\n\n# Suzieq -- Healthier Networks Through Network Observability\n\nWould you like to be able to easily answer trivial questions such as how many unique prefixes are there in your routing table, or how many MAC addresses are there in the MAC tables across the network? How about more difficult questions, such as what changes did your routing table see between 10 pm and midnight last night, or which of your nodes have been up the longest, or which BGP sessions have had the most routing updates? How about being able to answer if your OSPF (or BGP) sessions are working correctly, or is all well with your EVPN? How about a quick way to determine the amount of ECMP at every hop between two endpoints? Do you wish you could easily validate the configuration you deployed across your network?\n\nDo you login to every network node you have to figure out answers to a questions like these? Do you then struggle to piece the information together into a consistent whole across the various formats provided by various vendors? Do you wish you had an **open source, multi-vendor** tool that could help you answer questions like these and more?\n\nIf you answered yes to one or more of these questions, then Suzieq is a tool that we think will be interesting to you.  **Suzieq helps you find things in your network.**\n\n**Suzieq** is both a framework and an application using that framework, that is focused on **improving the observability of your network**.  We define observability as the ability of a system to answer either trivial or complex questions that you pose as you go about operating your network. How easily you can answer your questions is a measure of how good the system's observability is. A good observable system goes well beyond monitoring and alerting. Suzieq is primarily meant for use by network engineers and designers.\n\nSuzieq does multiple things. It [collects](https://suzieq.readthedocs.io/en/latest/poller/) data from devices and systems across your network. It normalizes the data and then stores it in a vendor independent way. Then it allows analysis of that data. With the applications that we build on top of the framework we want to demonstrate a different and more systematic approach to thinking about networks. We want to show how useful it is to think of your network holistically.\n\n## Quick Start\n\n### Using Docker Container\n\nWe want to make it as easy as possible for you to start engaging with Suzieq so\nwe have a demo that has data included in the image. To get started:\n\n* `docker run -it -p 8501:8501 --name suzieq netenglabs/suzieq-demo`\n* `suzieq-cli` for the CLI OR\n* `suzieq-gui` for the GUI. Connect to http://localhost:8501 via the browser to access the GUI\n\nWhen you're within the suzieq-cli, you can run ```device unique columns=namespace``` to see the list of different scenarios, we've gathered data for.\n\nAdditional information about running the analyzer (suzieq-cli) is available via\nthe official [documentation page](https://suzieq.readthedocs.io/en/latest/).\n\nTo start collecting data for your network, create an inventory file to gather the data from following the instructions [here](https://suzieq.readthedocs.io/en/latest/poller/). Decide the directory where the data will be stored (ensure you have sufficient available space if you're going to be running the poller, say 100 MB at least). Lets call this dbdir. Now launch the suzieq docker container as follows:\n\n* ```docker run -itd -v <parquet-out-local-dir>:/suzieq/parquet -v <inventory-file>:/suzieq/inventory.yml --name sq-poller netenglabs/suzieq:latest```\n* Connect to the container via ```docker attach sq-poller```\n* Launch the poller with the appropriate options. For example, ```sq-poller -D inventory.yml -n mydatacenter``` where mydatacenter is the name of the namespace where the data associated with the inventory is stored and inventory.yml is the inventory file in Suzieq poller native format (Use -a instead of -D if you're using Ansible inventory file format).\n\n### Using Python Packaging\n\nIf you don't want to use docker container or cannot use a docker container, an alternative approach is to install Suzieq as a python package. It is **strongly** recommended to install suzieq inside a virtual environment. If you already use a tool to create and manage virtual environments, you can skip the step of creating a virtual envirobment below.\n\nSuzieq requires python version 3.7.1 at least, and has been tested with python versions 3.7 and 3.8. It has not been tested to work on Windows. Use Linux (recommended) or macOS. To create a virtual environment, in case you haven't got a tool to create one, type:\n\n```bash\npython -m venv suzieq\n```\n\nThis creates a directory called suzieq and all suzieq related info is stored there. Switch to that directory and activate the virtual environment with:\n\n```bash\nsource activate\n```\n\nNow the virtual environment is alive and you can install suzieq. To install suzieq, execute:\n\n```bash\npip install suzieq\n```\n\nOnce the command completes, you have the main programs of suzieq available for use:\n\n* sq-poller: For polling the devices and gathering the data\n* suzieq-gui: For launching the GUI\n* suzieq-cli: For running the CLI\n* sq-rest-server: For running the REST API server\n\n[The official documentation is at suzieq.readthedocs.io](https://suzieq.readthedocs.io/en/latest/), and you can watch the screencasts about Suzieq on [Youtube](https://www.youtube.com/results?search_query=netenglabs).\n\n# Analysis\n\nSuzieq supports Analysis using CLI, GUI, REST API, and python objects. For the most part they are equivalent, though with the GUI we have combined the output of multiple commands of the CLI into one page.\n\nThe GUI has a status page to let you know what the status of entities in your network.\n![Suzieq GUI status](images/status.png)\n\nThe Xplore page lets you dive into what is in your network. ![Explore device](images/devices-gui.png)\n\nThe CLI supports the same kind of analysis as the explore page. ![CLI device](images/devices-cli.png)\n\n[More examples of the CLI can be seen in the docs and blog posts we've created.](https://suzieq.readthedocs.io/en/latest/analyzer/)\n\n## Path\n\nSuzieq has the ability to show the path between two IP addresses, including the ability to show the path through EVPN overlay. You can use this to see each of the paths from a source to a destination and to see if you have anything asymetrical in your paths. ![GUI PATH](images/path-gui.png)\n\n## Asserts\n\nOne of Suzieq's powerful capabilities are asserts, which are statements that should be true in the network. We've only just started on asserts; what Suzieq has now only demonstrates it's power, there's a lot more to be added in this space. ![interfaces assert](images/interfaces-assert.png)\n\n# Suzieq Data\n\n**Suzieq supports gathering data from Cumulus, EOS, IOS, IOSXE, IOSXR, JunOS(QFX, MX, EX, SRX supported), NXOS and SONIC routers, and Linux servers.** Suzieq gathers:\n\n* Basic device info including serial number, model, version, platform etc.\n* Interfaces\n* LLDP\n* MAC address table (VPLS MAC table for Junos MX)\n* MLAG\n* Routing table\n* ARP/ND table\n* OSPFv2\n* BGP\n* EVPN VNI info\n\nWe're adding support for more platforms and features with every release. See [the documentation](https://suzieq.readthedocs.io/en/latest/tables/) on details of specific tables and its NOS support.\n\nWe're also looking for collaborators to help us make Suzieq a truly useful multi-vendor, open source platform for observing all aspects of networking. Please read the [collaboration document](./CONTRIBUTING.md) for ideas on how you can help.\n\n# Release Notes\n\nThe official release notes are [here](https://suzieq.readthedocs.io/en/latest/release-notes/).\n\n# Engage\n\nYou can join the conversation via [slack](https://join.slack.com/t/netenglabs/shared_invite/zt-g64xa6lc-SeP2OAj~3uLbgOWJniLslA). Send email to suzieq AT stardustsystems.net with the email address to send the Slack invitation to.\n\n# Additional Documentation & Screencasts\n\nWe've done some blogging about Suzieq:\n\n* [Introducing Suzieq](https://elegantnetwork.github.io/posts/Suzieq/)\n* [10ish ways to explore your network with Suzieq](https://elegantnetwork.github.io/posts/10ish_ways_to_explore_your_network_with_Suzieq/)\n* [Questions to Suzieq](https://elegantnetwork.github.io/posts/10qa-suzieq/)\n* [Time in Suzieq](https://elegantnetwork.github.io/posts/time-suzieq/)\n\nWe've also been adding screencasts on [Youtube](https://www.youtube.com/results?search_query=netenglabs).\n\n# Suzieq Priorities\n\nWe don't have a roadmap, but we do have a list of our [priorities](https://github.com/netenglabs/suzieq/blob/master/docs/2020-priority.md). We mix this with the [issues reported](https://github.com/netenglabs/suzieq/issues).\n",
    'author': 'suzieq dev team',
    'author_email': None,
    'maintainer': 'suzieq dev team',
    'maintainer_email': None,
    'url': 'https://www.stardustsystems.net/suzieq/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>3.7.1,<3.9',
}


setup(**setup_kwargs)



from typing import Optional, Sequence, List
import os
import argparse
import json
import sys
from enum import Enum
from fastapi import FastAPI, HTTPException, Query, Depends, Security, Request
from fastapi.security.api_key import APIKeyQuery, APIKeyHeader
from fastapi.responses import Response
from starlette import status
import uuid
import inspect
import logging
import uvicorn

from suzieq.sqobjects import get_sqobject
from suzieq.utils import (load_sq_config, get_sq_install_dir, get_log_params,
                          sq_get_config_file)

API_KEY_NAME = 'access_token'

api_key_query = APIKeyQuery(name=API_KEY_NAME, auto_error=False)
api_key_header = APIKeyHeader(name=API_KEY_NAME, auto_error=False)


def check_config_file():
    if not getattr(app, 'cfg_file', None):
        print('missing config file')
        sys.exit(1)


# Changing the default URLs to help with reverse proxy stuff as described
# in issue #381 (https://github.com/netenglabs/suzieq/issues/381)
app = FastAPI(on_startup=[check_config_file],
              openapi_url="/api/openapi.json",
              docs_url="/api/docs",
              redoc_url="/api/redoc")


def app_init(cfg_file):
    '''This is the actual API initilaizer'''
    global app

    app.cfg_file = cfg_file

    return app


def get_configured_api_key():
    cfg = load_sq_config(config_file=app.cfg_file)
    try:
        api_key = cfg['rest']['API_KEY']
    except KeyError:
        print('missing API_KEY in config file')
        sys.exit(1)

    return api_key


async def get_api_key(api_key_query: str = Security(api_key_query),
                      api_key_header: str = Security(api_key_header)):

    api_key = get_configured_api_key()
    if api_key_query == api_key:
        return api_key_query
    elif api_key_header == api_key:
        return api_key_header
    else:

        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API Key",
        )


def get_cert_files(cfg):
    sqdir = get_sq_install_dir()
    ssl_certfile = cfg.get('rest', {}) \
                      .get('rest-certfile', f'{sqdir}/config/etc/cert.pem')

    ssl_keyfile = cfg.get('rest', {}) \
                     .get('rest-keyfile', f'{sqdir}/config/etc/key.pem')

    if not os.path.isfile(ssl_certfile):
        print(f"ERROR: Missing certificate file: {ssl_certfile}")
        sys.exit(1)

    if not os.path.isfile(ssl_keyfile):
        print(f"ERROR: Missing certificate file: {ssl_keyfile}")
        sys.exit(1)

    return ssl_keyfile,  ssl_certfile


def get_log_config_level(cfg):

    logfile, loglevel, logsize, log_stdout = get_log_params(
        'rest', cfg, '/tmp/sq-rest-server.log')

    log_config = uvicorn.config.LOGGING_CONFIG
    if logfile and not log_stdout:
        log_config['handlers']['access']['filename'] = logfile
        log_config['handlers']['access']['class'] = 'logging.handlers.RotatingFileHandler'
        log_config['handlers']['access']['maxBytes'] = logsize
        log_config['handlers']['access']['backupCount'] = 2

        log_config['handlers']['default']['class'] = 'logging.handlers.RotatingFileHandler'
        log_config['handlers']['default']['maxBytes'] = logsize
        log_config['handlers']['default']['backupCount'] = 2
        log_config['handlers']['default']['filename'] = logfile

        if 'stream' in log_config['handlers']['default']:
            del log_config['handlers']['default']['stream']
            del log_config['handlers']['access']['stream']

    return log_config, loglevel


def rest_main(*args) -> None:
    """The main function for the REST server

    Args:
        config_file (str): The suzieq config file
        no_https (bool): If true, disable https
    """

    if not args:
        args = sys.argv

    parser = argparse.ArgumentParser(args)
    parser.add_argument(
        "-c",
        "--config",
        type=str, help="alternate config file",
        default=None
    )
    parser.add_argument(
        "--no-https",
        help="Turn off HTTPS",
        default=False, action='store_true',
    )
    userargs = parser.parse_args()

    config_file = sq_get_config_file(userargs.config)
    app = app_init(config_file)
    cfg = load_sq_config(config_file=config_file)
    try:
        api_key = cfg['rest']['API_KEY']
    except KeyError:
        print('missing API_KEY in config file')
        exit(1)

    logcfg, loglevel = get_log_config_level(cfg)

    no_https = cfg.get('rest', {}).get('no-https', False) or userargs.no_https

    srvr_addr = cfg.get('rest', {}).get('address', '127.0.0.1')
    srvr_port = cfg.get('rest', {}).get('port', 8000)

    if no_https:
        uvicorn.run(app, host=srvr_addr, port=srvr_port,
                    )
    else:
        ssl_keyfile, ssl_certfile = get_cert_files(cfg)
        uvicorn.run(app, host=srvr_addr, port=srvr_port,
                    ssl_keyfile=ssl_keyfile,
                    ssl_certfile=ssl_certfile)


"""
each of these read functions behaves the same, it gets the arguments
puts them into dicts and passes them to sqobjects

assume that all API functions are named read_*
"""


class CommonVerbs(str, Enum):
    show = "show"
    summarize = "summarize"
    unique = "unique"
    top = "top"


class RouteVerbs(str, Enum):
    show = "show"
    summarize = "summarize"
    unique = "unique"
    lpm = "lpm"
    top = "top"


class TableVerbs(str, Enum):
    show = "show"
    summarize = "summarize"
    unique = "unique"
    top = "top"
    describe = "describe"


class DeviceStatus(str, Enum):
    alive = "alive"
    dead = "dead"
    neverpoll = "neverpoll"


class BgpStateValues(str, Enum):
    ESTABLISHED = "Established"
    NOTESTD = "NotEstd"


class IfStateValues(str, Enum):
    UP = "up"
    DOWN = "down"
    ERRDISABLED = "errDisabled"
    NOTCONNECTED = "notConnected"


class OspfStateValues(str, Enum):
    FULL = "full"
    PASSIVE = "passive"
    OTHER = "other"


class ViewValues(str, Enum):
    latest = "latest"
    all = "all"
    changes = "changes"


class AssertStatusValues(str, Enum):
    PASS = "pass"
    FAIL = "fail"
    ALL = "all"


class InventoryStatusValues(str, Enum):
    PRESENT = "present"
    ABSENT = "absent"


# The logic in the code below is that you have a common function to
# split the common arguments across all the functions, and split the
# object-specific arguments in the object function itself.
@app.get("/api/v1/{rest_of_path:path}", deprecated=True)
async def deprecated_function(request: Request, rest_of_path: str):
    return([{'error': 'v1 is deprecated, use API version v2'}])


@app.get("/api/v2/address/{verb}")
async def query_address(verb: CommonVerbs, request: Request,
                        token: str = Depends(get_api_key),
                        format: str = None,
                        hostname: List[str] = Query(None),
                        start_time: str = "", end_time: str = "",
                        view: ViewValues = "latest",
                        namespace: List[str] = Query(None),
                        columns: List[str] = Query(default=["default"]),
                        address: List[str] = Query(None),
                        ipvers: str = None, what: str = None,
                        vrf: str = None, query_str: str = None,
                        ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/arpnd/{verb}")
async def query_arpnd(verb: CommonVerbs, request: Request,
                      token: str = Depends(get_api_key),
                      format: str = None,
                      hostname: List[str] = Query(None),
                      start_time: str = "", end_time: str = "",
                      view: ViewValues = "latest",
                      namespace: List[str] = Query(None),
                      columns: List[str] = Query(default=["default"]),
                      ipAddress: List[str] = Query(None),
                      macaddr: List[str] = Query(None),
                      oif: List[str] = Query(None),
                      query_str: str = None, what: str = None,
                      ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/bgp/assert")
async def query_bgp_assert(request: Request,
                           token: str = Depends(get_api_key),
                           format: str = None,
                           hostname: List[str] = Query(None),
                           start_time: str = "", end_time: str = "",
                           view: ViewValues = "latest",
                           namespace: List[str] = Query(None),
                           columns: List[str] = Query(default=["default"]),
                           vrf: List[str] = Query(None),
                           status: AssertStatusValues = Query(None),
                           query_str: str = None,
                           ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, "assert", request, locals())


@ app.get("/api/v2/bgp/{verb}")
async def query_bgp(verb: CommonVerbs, request: Request,
                    token: str = Depends(get_api_key),
                    format: str = None,
                    hostname: List[str] = Query(None),
                    start_time: str = "", end_time: str = "",
                    view: ViewValues = "latest",
                    namespace: List[str] = Query(None),
                    columns: List[str] = Query(default=["default"]),
                    peer: List[str] = Query(None),
                    state: BgpStateValues = Query(None),
                    vrf: List[str] = Query(None),
                    asn: List[str] = Query(None),
                    status: AssertStatusValues = Query(None),
                    query_str: str = None, what: str = None,
                    ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/device/{verb}")
async def query_device(verb: CommonVerbs, request: Request,
                       token: str = Depends(get_api_key),
                       format: str = None,
                       hostname: List[str] = Query(None),
                       start_time: str = "", end_time: str = "",
                       view: ViewValues = "latest",
                       namespace: List[str] = Query(None),
                       columns: List[str] = Query(default=["default"]),
                       query_str: str = None,
                       os: List[str] = Query(None),
                       vendor: List[str] = Query(None),
                       model: List[str] = Query(None),
                       version: str = "", what: str = None,
                       status: List[DeviceStatus] = Query(None),
                       ):
    function_name = inspect.currentframe().f_code.co_name
    if status:
        status = [x.value for x in status]  # convert enum to string
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/devconfig/{verb}")
async def query_devconfig(verb: CommonVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          view: ViewValues = "latest",
                          namespace: List[str] = Query(None),
                          columns: List[str] = Query(default=["default"]),
                          query_str: str = None,
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/evpnVni/assert")
async def query_evpnVni_assert(request: Request,
                               token: str = Depends(get_api_key),
                               format: str = None,
                               hostname: List[str] = Query(None),
                               start_time: str = "", end_time: str = "",
                               view: ViewValues = "latest",
                               namespace: List[str] = Query(None),
                               columns: List[str] = Query(default=["default"]),
                               status: AssertStatusValues = None,
                               query_str: str = None,
                               ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, "assert", request, locals())


@ app.get("/api/v2/evpnVni/{verb}")
async def query_evpnVni(verb: CommonVerbs, request: Request,
                        token: str = Depends(get_api_key),
                        format: str = None,
                        hostname: List[str] = Query(None),
                        start_time: str = "", end_time: str = "",
                        view: ViewValues = "latest",
                        namespace: List[str] = Query(None),
                        columns: List[str] = Query(default=["default"]),
                        vni: List[str] = Query(None),
                        priVtepIp: List[str] = Query(None),
                        status: AssertStatusValues = None,
                        query_str: str = None, what: str = None,
                        ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/fs/{verb}")
async def query_fs(verb: CommonVerbs, request: Request,
                   token: str = Depends(get_api_key),
                   format: str = None,
                   hostname: List[str] = Query(None),
                   start_time: str = "", end_time: str = "",
                   view: ViewValues = "latest",
                   namespace: List[str] = Query(None),
                   columns: List[str] = Query(default=["default"]),
                   mountPoint: List[str] = Query(None), what: str = None,
                   usedPercent: str = None, query_str: str = None,
                   ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/interface/assert")
async def query_interface_assert(request: Request,
                                 token: str = Depends(get_api_key),
                                 format: str = None,
                                 hostname: List[str] = Query(None),
                                 start_time: str = "", end_time: str = "",
                                 view: ViewValues = "latest",
                                 namespace: List[str] = Query(None),
                                 columns: List[str] = Query(default=["default"]),
                                 ifname: List[str] = Query(None),
                                 what: str = None,
                                 matchval: int = Query(None, alias="value"),
                                 status: AssertStatusValues = Query(None),
                                 ignore_missing_peer: bool = Query(False),
                                 query_str: str = None,
                                 ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, "assert", request, locals())


@ app.get("/api/v2/interface/{verb}")
async def query_interface(verb: CommonVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          view: ViewValues = "latest",
                          namespace: List[str] = Query(None),
                          columns: List[str] = Query(default=["default"]),
                          ifname: List[str] = Query(None),
                          state: IfStateValues = Query(None),
                          type: List[str] = Query(None),
                          what: str = None, vrf: List[str] = Query(None),
                          master: List[str] = Query(None),
                          mtu: List[str] = Query(None),
                          ifindex: List[str] = Query(None),
                          matchval: int = Query(None, alias="value"),
                          status: AssertStatusValues = Query(None),
                          ignore_missing_peer: bool = Query(False),
                          query_str: str = None,
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/inventory/{verb}")
async def query_inventory(verb: CommonVerbs, request: Request,
                          token: str = Depends(get_api_key),
                          format: str = None,
                          hostname: List[str] = Query(None),
                          start_time: str = "", end_time: str = "",
                          view: ViewValues = "latest",
                          namespace: List[str] = Query(None),
                          columns: List[str] = Query(default=["default"]),
                          query_str: str = None,
                          type: List[str] = Query(None),
                          serial: List[str] = Query(None),
                          model: List[str] = Query(None),
                          vendor: List[str] = Query(None), what: str = None,
                          status: InventoryStatusValues = Query(None),
                          ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/lldp/{verb}")
async def query_lldp(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     ifname: List[str] = Query(None),
                     query_str: str = None, what: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/mac/{verb}")
async def query_mac(verb: CommonVerbs, request: Request,
                    token: str = Depends(get_api_key),
                    format: str = None,
                    hostname: List[str] = Query(None),
                    start_time: str = "", end_time: str = "",
                    view: ViewValues = "latest",
                    namespace: List[str] = Query(None),
                    columns: List[str] = Query(default=["default"]),
                    bd: str = None,
                    localOnly: str = None,
                    macaddr: List[str] = Query(None),
                    remoteVtepIp: List[str] = Query(None),
                    vlan: List[str] = Query(None),
                    query_str: str = None, what: str = None,
                    moveCount: str = None,
                    ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/mlag/{verb}")
async def query_mlag(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     query_str: str = None, what: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@app.get("/api/v2/network/find")
async def query_network_find(request: Request,
                             token: str = Depends(get_api_key),
                             format: str = None,
                             columns: List[str] = Query(default=["default"]),
                             namespace: List[str] = Query(None),
                             hostname: List[str] = Query(None),
                             start_time: str = "", end_time: str = "",
                             view: ViewValues = "latest",
                             address: str = "", vlan: str = '', vrf: str = '',
                             query_str: str = None,
                             ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, "find", request, locals())


@app.get("/api/v2/network/{verb}")
async def query_network(verb: CommonVerbs, request: Request,
                        token: str = Depends(get_api_key),
                        format: str = None,
                        columns: List[str] = Query(default=["default"]),
                        namespace: List[str] = Query(None),
                        hostname: List[str] = Query(None),
                        start_time: str = "", end_time: str = "",
                        version: str = "",
                        view: ViewValues = "latest",
                        model: List[str] = Query(None),
                        vendor: List[str] = Query(None),
                        os: List[str] = Query(None),
                        query_str: str = None, what: str = None,
                        ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/ospf/assert")
async def query_ospf_assert(request: Request,
                            token: str = Depends(get_api_key),
                            format: str = None,
                            hostname: List[str] = Query(None),
                            start_time: str = "", end_time: str = "",
                            view: ViewValues = "latest",
                            namespace: List[str] = Query(None),
                            columns: List[str] = Query(default=["default"]),
                            vrf: List[str] = Query(None),
                            status: AssertStatusValues = None,
                            query_str: str = None,
                            ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, "assert", request, locals())


@ app.get("/api/v2/ospf/{verb}")
async def query_ospf(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     ifname: List[str] = Query(None),
                     state: OspfStateValues = Query(None),
                     vrf: List[str] = Query(None),
                     status: AssertStatusValues = None,
                     query_str: str = None, what: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/path/{verb}")
async def query_path(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     vrf: str = None,
                     dest: str = None,
                     source: str = Query(None, alias="src"),
                     query_str: str = None, what: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/route/{verb}")
async def query_route(verb: RouteVerbs, request: Request,
                      token: str = Depends(get_api_key),
                      format: str = None,
                      hostname: List[str] = Query(None),
                      start_time: str = "", end_time: str = "",
                      view: ViewValues = "latest",
                      namespace: List[str] = Query(None),
                      columns: List[str] = Query(default=["default"]),
                      prefix: List[str] = Query(None),
                      vrf: List[str] = Query(None),
                      protocol: List[str] = Query(None),
                      prefixlen: str = None, ipvers: str = None,
                      add_filter: str = None, address: str = None,
                      query_str: str = None, what: str = None,
                      ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/sqPoller/{verb}")
async def query_sqPoller(verb: CommonVerbs, request: Request,
                         token: str = Depends(get_api_key),
                         format: str = None,
                         hostname: List[str] = Query(None),
                         start_time: str = "", end_time: str = "",
                         view: ViewValues = "latest",
                         namespace: List[str] = Query(None),
                         columns: List[str] = Query(default=["default"]),
                         service: str = None,
                         status: AssertStatusValues = Query(None),
                         query_str: str = None, what: str = None,
                         ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/topology/summarize")
async def query_topology_summarize(request: Request,
                                   token: str = Depends(get_api_key),
                                   format: str = None,
                                   hostname: List[str] = Query(None),
                                   start_time: str = "", end_time: str = "",
                                   view: ViewValues = "latest",
                                   namespace: List[str] = Query(None),
                                   columns: List[str] = Query(default=["default"]),
                                   via: List[str] = Query(None),
                                   query_str: str = None,
                                   ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, "summarize", request, locals())


@ app.get("/api/v2/topology/{verb}")
async def query_topology(verb: CommonVerbs, request: Request,
                         token: str = Depends(get_api_key),
                         format: str = None,
                         hostname: List[str] = Query(None),
                         start_time: str = "", end_time: str = "",
                         view: ViewValues = "latest",
                         namespace: List[str] = Query(None),
                         columns: List[str] = Query(default=["default"]),
                         polled: str = None,
                         via: List[str] = Query(None),
                         ifname: List[str] = Query(None),
                         peerHostname: List[str] = Query(None),
                         query_str: str = None, what: str = None,
                         ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/vlan/{verb}")
async def query_vlan(verb: CommonVerbs, request: Request,
                     token: str = Depends(get_api_key),
                     format: str = None,
                     hostname: List[str] = Query(None),
                     start_time: str = "", end_time: str = "",
                     view: ViewValues = "latest",
                     namespace: List[str] = Query(None),
                     columns: List[str] = Query(default=["default"]),
                     vlan: List[str] = Query(None),
                     state: str = None,
                     vlanName: List[str] = Query(None),
                     query_str: str = None, what: str = None,
                     ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


@ app.get("/api/v2/table/{verb}")
async def query_table(verb: TableVerbs, request: Request,
                      token: str = Depends(get_api_key),
                      format: str = None,
                      hostname: List[str] = Query(None),
                      start_time: str = "", end_time: str = "",
                      view: ViewValues = "latest", namespace: List[str] = Query(None),
                      columns: List[str] = Query(default=["default"]),
                      query_str: str = None
                      ):
    function_name = inspect.currentframe().f_code.co_name
    return read_shared(function_name, verb, request, locals())


def read_shared(function_name, verb, request, local_variables=None):
    """all the shared code for each of thse read functions"""

    command = function_name.split('_')[1]  # assumes fn name is query_<command>
    command_args, verb_args = create_filters(function_name, command, request,
                                             local_variables)

    verb = cleanup_verb(verb)

    columns = local_variables.get('columns', None)
    format = local_variables.get('format', None)
    ret, svc_inst = run_command_verb(
        command, verb, command_args, verb_args, columns, format)

    return ret


def create_filters(function_name, command, request, local_vars):
    command_args = {}
    verb_args = {}
    remove_args = ['verb', 'token', 'format', 'request', 'access_token']
    all_cmd_args = ['namespace', 'hostname',
                    'start_time', 'end_time', 'view', 'columns', 'format']
    both_verb_and_command = ['namespace', 'hostname', ]
    alias_args = {'src': 'source'}
    def_val_args = {
        'namespace': [],
        'hostname': [],
    }

    query_ks = request.query_params
    for arg in query_ks.keys():
        if arg in remove_args:
            continue
        if arg in alias_args:
            tryarg = alias_args[arg]
        else:
            tryarg = arg
        if arg in all_cmd_args:
            if query_ks.get(arg) is not None:
                command_args[tryarg] = local_vars.get(tryarg, None)
                if arg in both_verb_and_command:
                    verb_args[tryarg] = command_args[arg]
        else:
            if query_ks.get(arg) is not None:
                verb_args[tryarg] = local_vars.get(tryarg, None)

    return command_args, verb_args


def cleanup_verb(verb):
    if verb == 'show':
        verb = 'get'
    if verb == 'assert':
        verb = 'aver'
    return verb


def create_command_args(hostname='', start_time='', end_time='', view='latest',
                        namespace='', columns='default'):
    command_args = {'hostname': hostname,
                    'start_time': start_time,
                    'end_time': end_time,
                    'view': view,
                    'namespace': namespace,
                    'columns': columns}
    return command_args


def get_svc(command):
    """based on the command, find the module and service that the command is in
    return the service
    """
    command_name = command

    svc = get_sqobject(command_name)
    return svc


def run_command_verb(command, verb, command_args, verb_args, columns=['default'], format=None):
    """
    Runs the command and verb with the command_args and verb_args as dictionaries

    HTTP Return Codes
        404 -- Missing command or argument (including missing valid path)
        405 -- Missing or incorrect query parameters
        422 -- FastAPI validation errors
        500 -- Exceptions
    """
    svc = get_svc(command)
    try:
        svc_inst = svc(**command_args, config_file=app.cfg_file)
        df = getattr(svc_inst, verb)(**verb_args)

    except AttributeError as err:
        return_error(
            404, f"{verb} not supported for {command} or missing arguement: {err}")

    except NotImplementedError as err:
        return_error(404, f"{verb} not supported for {command}: {err}")

    except TypeError as err:
        return_error(405, f"bad keyword/filter for {command} {verb}: {err}")

    except ValueError as err:
        return_error(405, f"bad keyword/filter for {command} {verb}: {err}")

    except Exception as err:
        return_error(
            500, f"exceptional exception {verb} for {command} of type {type(err)}: {err}")

    if df.columns.to_list() == ['error']:
        return_error(
            405, f"bad keyword/filter for {command} {verb}: {df['error'][0]}")

    if columns != ['default'] and columns != ['*'] and columns is not None:
        df = df[columns]

    if format == 'markdown':
        # have to return a Reponse so that it won't turn the markdown into JSON
        return Response(content=df.to_markdown()), svc_inst

    if verb == 'summarize':
        json_orient = 'columns'
    else:
        json_orient = 'records'
    return json.loads(df.to_json(orient=json_orient)), svc_inst


def return_error(code: int, msg: str):
    u = uuid.uuid1()
    msg = f"{msg} id={u}"
    logger = logging.getLogger('uvicorn')
    logger.info(msg)
    raise HTTPException(status_code=code, detail=msg)


@ app.get("/api/v2/{command}", include_in_schema=False)
def missing_verb(command):
    return_error(
        404, f'{command} command missing a verb. for example '
        f'/api/v2/{command}/show')


@ app.get("/", include_in_schema=False)
def bad_path():
    return_error(
        404, "bad path. Try something like '/api/v2/device/show' or '/api/docs'")



import typing
import os
from nubia import command, argument, context, CompletionDataSource
from nubia.internal.commands.help import HelpCommand
from nubia.internal.cmdbase import Command, AutoCommand
from prompt_toolkit.completion import Completion
from nubia.internal import parser
from termcolor import cprint, colored
from suzieq.sqobjects import get_tables


@command("set")
@argument("namespace", description="namespace to qualify selection")
@argument("hostname", description="Name of host to qualify selection")
@argument(
    "start_time", description="Start of time window in YYYY-MM-dd HH:mm:SS format"
)
@argument("end_time", description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument("pager", description="Enable pagination prompt on longer outputs",
          choices=['on', 'off'])
@argument('col_width', description='Max Width of each column in table display')
@argument(
    "engine",
    choices=["pandas"],
    description="Use Pandas for non-SQL commands",
)
@argument(
    "datadir",
    description="Set the data directory for the command"
)
def set_ctxt(
        pager: str = 'on',
        hostname: typing.List[str] = [],
        start_time: str = "",
        end_time: str = "",
        namespace: typing.List[str] = [],
        engine: str = "",
        datadir: str = "",
        col_width: int = 50,
):
    """set certain contexts for subsequent commands. Cmd is additive"""
    plugin_ctx = context.get_context()

    if namespace:
        plugin_ctx.namespace = namespace

    if hostname:
        plugin_ctx.hostname = hostname

    if start_time:
        plugin_ctx.start_time = start_time

    if end_time:
        plugin_ctx.end_time = end_time

    if engine:
        plugin_ctx.change_engine(engine)

    if datadir:
        if os.path.isdir(datadir):
            plugin_ctx.cfg['data-directory'] = datadir
        else:
            print(f'{datadir} is not a valid directory')

    if col_width:
        plugin_ctx.col_width = int(col_width)

    if pager == 'on':
        plugin_ctx.pager = True


@command("clear")
@argument("namespace", description="namespace to qualify selection")
@argument("hostname", description="Name of host to qualify selection")
@argument(
    "start_time", description="Start of time window in YYYY-MM-dd HH:mm:SS format"
)
@argument("end_time", description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument("pager", description="End of time window in YYYY-MM-dd HH:mm:SS format")
def clear_ctxt(
        pager: str = 'off',
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        namespace: str = "",
):
    """clear certain contexts for subsequent commands. Cmd is additive"""
    plugin_ctx = context.get_context()

    if namespace:
        plugin_ctx.namespace = []

    if hostname:
        plugin_ctx.hostname = []

    if start_time:
        plugin_ctx.start_time = ""

    if end_time:
        plugin_ctx.end_time = ""

    if pager:
        plugin_ctx.pager = False


class SqHelpCommand (Command):
    HELP = 'Get help on a command'
    thiscmd = ["help", "?"]

    def __init__(self):
        super().__init__()
        self._allcmds = self._sqcmds = None

    def run_interactive(self, cmd, args, raw):
        arglist = args.split()
        return self._help_cmd(*arglist)

    def get_command_names(self):
        return self.thiscmd

    def add_arguments(self, parser):
        parser.add_parser("help")

    def get_help(self, cmd, *args):
        return self.thiscmd[0]

    def get_completions(self, cmd, document, complete_event):
        exploded = document.text.lstrip().split(" ", 1)
        self._build_cmd_verb_list()
        if len(exploded) <= 1:
            if not document.text:
                return [Completion(x)
                        for x in self._sqcmds]
            completions = [Completion(text=x, start_position=-len(document.text))
                           for x in self._sqcmds
                           if x.startswith(document.text)]
            return completions

        service = exploded[0].lower()
        verb = exploded[1]

        verbs = [x[0].replace('aver', 'assert')
                 for x in self._allcmds[service].metadata.subcommands
                 if service in self._sqcmds]
        if not verb:
            return [Completion(x) for x in verbs]
        completions = [Completion(text=x, start_position=-len(verb))
                       for x in sorted(verbs)
                       if x.startswith(verb)]
        return completions

    def _help_cmd(self, *args):
        """Show help for a service"""
        if len(args) == 0:
            helpcmd = HelpCommand()
            helpcmd.run_interactive('', '', '')
            cprint("Use "
                   f"{colored('help <name of command/service> [<verb>]', 'cyan')} "
                   "to get more help")
            cprint(f"For example: {colored('help route', 'cyan')}"
                   f" or {colored('help route show', 'cyan')}")
            return
        if len(args) == 1:
            service = args[0]
            verb = ''
        else:
            service = args[0]
            verb = args[1]

        self._build_cmd_verb_list()
        if service in self._sqcmds:
            if verb:
                self._allcmds[service].run_interactive(
                    service, f'help command={verb}', '')
            else:
                self._allcmds[service].run_interactive(service, f'help', '')

    def _build_cmd_verb_list(self):
        if not self._allcmds:
            ctx = context.get_context()
            self._allcmds = ctx.registry.get_all_commands_map()
            self._sqcmds = [x for x in sorted(self._allcmds)
                            if not self._allcmds[x].built_in]



import re
import numpy as np

from suzieq.poller.services.service import Service
from suzieq.utils import (convert_rangestring_to_list,
                          convert_macaddr_format_to_colon)


class EvpnVniService(Service):
    """evpnVni service. Different class because output needs to be munged"""

    def clean_json_input(self, data):
        """evpnVni JSON output is busted across many NOS. Fix it"""

        devtype = data.get("devtype", None)
        if any(x == devtype for x in ["cumulus", "sonic", "linux"]):
            data['data'] = '[' + re.sub(r'}\n\n{\n', r'},\n\n{\n',
                                        data['data']) + ']'
            return data['data']
        elif devtype.startswith('junos'):
            data['data'] = data['data'].replace('}, \n    }\n', '} \n    }\n')
            return data['data']
        elif devtype == "eos":
            if data.get('cmd', '').startswith('show interfaces Vxlan $'):
                if data['data'].startswith('%'):
                    data['data'] = '{}'
                    return data['data']

    def _clean_eos_data(self, processed_data, raw_data):
        new_entries = []

        if not processed_data:
            return processed_data

        for entry in processed_data:
            vni2vrfmap = {}
            for vrf in entry['_vrf2VniMap']:
                vni2vrfmap[entry['_vrf2VniMap'][vrf]] = vrf

            vtepMap = entry.get('_vlan2VtepMap', {})
            replType = entry.get('replicationType')
            if replType == 'headendVcs':
                replType = 'ingressBGP'
            elif entry.get('mcastGroup', '') != "0.0.0.0":
                replType = "multicast"
            else:
                replType = ''
            for vlan in entry['_vlan2VniMap']:
                new_entry = {}
                vni = entry['_vlan2VniMap'][vlan].get('vni', 0)
                new_entry['vni'] = vni
                new_entry['vrf'] = vni2vrfmap.get(vni, '')
                new_entry['state'] = entry['state']
                new_entry['ifname'] = entry['ifname']
                new_entry['vlan'] = vlan
                new_entry['priVtepIp'] = entry['priVtepIp']
                vteplist = vtepMap.get(vlan, {})
                vteplist = (vteplist.get('remoteVtepAddr', []) +
                            vteplist.get('remoteVtepAddr6', []))
                new_entry['remoteVtepList'] = vteplist
                new_entry['replicationType'] = replType
                new_entry['mcastGroup'] = entry['mcastGroup']
                if new_entry['vrf']:
                    new_entry['type'] = 'L3'
                else:
                    new_entry['type'] = 'L2'
                new_entry['ifname'] = entry.get('ifname', '')

                new_entries.append(new_entry)

        processed_data = new_entries
        return processed_data

    def _clean_cumulus_data(self, processed_data, raw_data):
        """Clean out null entries among other cleanup"""

        del_indices = []
        for i, entry in enumerate(processed_data):
            if entry['vni'] is None:
                del_indices.append(i)
            if entry['mcastGroup'] and entry['mcastGroup'] != "0.0.0.0":
                entry['replicationType'] = 'multicast'
            elif entry['type'] != 'L3':
                entry['replicationType'] = 'ingressBGP'
                entry['mcastGroup'] = "0.0.0.0"
            else:
                entry['replicationType'] = ''
                entry['mcastGroup'] = "0.0.0.0"
                entry['remoteVtepList'] = None

            entry['state'] = entry.get('state', 'up').lower()
            entry['l2VniList'] = set(entry['l2VniList'])
        processed_data = np.delete(processed_data, del_indices).tolist()

        return processed_data

    def _clean_nxos_data(self, processed_data, raw_data):
        """Merge peer records with VNI records to yield VNI-based records"""

        vni_dict = {}
        drop_indices = []

        for i, entry in enumerate(processed_data):
            if not entry['vni']:
                drop_indices.append(i)
                continue

            if entry['_entryType'] == 'VNI':
                type, vrf = entry['type'].split()
                if type == 'L3':
                    entry['vrf'] = vrf[1:-1]  # strip off '[' and ']'
                entry['type'] = type
                if 'sviState' in entry:
                    entry['state'] = entry['sviState'].split()[0].lower()
                if re.search(r'[0-9.]+', entry.get('replicationType', '')):
                    entry['mcastGroup'] = entry['replicationType']
                    entry['replicationType'] = 'multicast'
                elif entry['type'] != 'L3':
                    entry['replicationType'] = 'ingressBGP'
                    entry['mcastGroup'] = "0.0.0.0"
                else:
                    entry['replicationType'] = ''
                    entry['mcastGroup'] = "0.0.0.0"

                # we'll fill this with the peers entries
                entry['remoteVtepList'] = []
                entry['state'] = entry['state'].lower()
                entry['vlan'] = int(entry['vlan'])
                vni_dict[entry['vni']] = entry

            elif entry['_entryType'] == 'peers':
                vni_list = convert_rangestring_to_list(
                    entry.get('_vniList', ''))
                for vni in vni_list:
                    vni_entry = vni_dict.get(vni, None)
                    if vni_entry:
                        vni_entry['remoteVtepList'].append(entry['vni'])
                drop_indices.append(i)

            elif entry['_entryType'] == 'iface':
                if entry.get('encapType', '') != "VXLAN":
                    continue

                for vni in vni_dict:
                    if vni_dict[vni]['ifname'] != entry['ifname']:
                        continue
                    vni_dict[vni]['priVtepIp'] = entry.get('priVtepIp', '')
                    secIP = entry.get('secVtepIp', '')
                    if secIP == '0.0.0.0':
                        secIP = ''
                    vni_dict[vni]['secVtepIp'] = secIP
                    vni_dict[vni]['routerMac'] = convert_macaddr_format_to_colon(
                        entry.get('routerMac', '00:00:00:00:00:00'))

                drop_indices.append(i)

        processed_data = np.delete(processed_data, drop_indices).tolist()

        return processed_data

    def _clean_junos_data(self, processed_data, raw_data):

        newntries = {}

        for entry in processed_data:
            if entry['_entryType'] == 'instance':
                if entry['_vniList'] is None:
                    continue
                for i, vni in enumerate(entry['_vniList']):
                    irb_iflist = entry.get('_irbIfList', [])
                    vrflist = entry.get('_vrfList', [])
                    vlan = entry['_vlanList'][i]
                    irbif = f'irb.{vlan}'
                    try:
                        index = irb_iflist.index(irbif)
                        vrf = vrflist[index]
                    except ValueError:
                        vrf = ''
                    except IndexError:
                        vrf = ''

                    if vni not in newntries:
                        vni_entry = {
                            'vni': int(vni),
                            'remoteVtepList': [],
                            'type': 'L2',
                            'state': 'up',
                            'vlan': int(vlan),
                            'numRemoteVteps': 0,
                            'numMacs': 0,
                            'numArpNd': 0,
                            'vrf': vrf,
                            'os': 'junos'
                        }
                    newntries[vni] = vni_entry
                    continue
            elif entry['_entryType'] == 'l3':
                vni = int(entry.get('vni', '0'))
                priVtepIp = entry.get('priVtepIp', '')

                if not priVtepIp and not vni:
                    continue

                vni_entry = {
                    'vni': vni,
                    'remoteVtepList': [],
                    'priVtepIp': priVtepIp,
                    'type': 'L3',
                    'state': 'up',
                    'numRemoteVteps': 0,
                    'routerMac': entry['routerMac'],
                    'numMacs': 0,
                    'numArpNd': 0,
                    'mcastGroup': '0.0.0.0',
                    'vrf': entry['vrf'],
                    'os': 'junos'
                }
                # Add the primary VTEP IP into the L2 entries as well
                for l2vni in newntries:
                    newntries[l2vni]['priVtepIp'] = priVtepIp

                newntries[vni] = vni_entry
                continue
            elif entry['_entryType'] == 'remote':
                priVtepIp = entry.get('priVtepIp', '[{"data": ""}]')[0]['data']
                for i, vni in enumerate(entry.get('_vniList', [])):
                    vni_entry = newntries.get(vni, {})
                    if not vni_entry:
                        vni_entry = {
                            'vni': int(vni),
                            'remoteVtepList': [],
                            'priVtepIp': priVtepIp,
                            'type': 'L2',
                            'state': 'up',
                            'numRemoteVteps': len(entry['_floodVtepList']),
                            'numMacs': 0,
                            'numArpNd': 0,
                            'os': 'junos'
                        }
                        newntries[vni] = vni_entry

                    vni_entry['priVtepIp'] = priVtepIp
                    if entry['replicationType'][i] == '0.0.0.0':
                        vni_entry['replicationType'] = 'ingressBGP'
                        vni_entry['mcastGroup'] = "0.0.0.0"
                    else:
                        vni_entry['replicationType'] = 'multicast'
                        vni_entry['mcastGroup'] = entry['replicationType'][i]

                    vni_entry['remoteVtepList'].append(
                        entry.get('_floodVtepList', ''))

        processed_data = list(newntries.values())
        return processed_data

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_cumulus_data(processed_data, raw_data)

    def _clean_linux_data(self, processed_data, raw_data):
        return self._clean_cumulus_data(processed_data, raw_data)



from streamlit.hashing import _CodeHasher
from streamlit.report_thread import get_report_ctx
from streamlit.server.server import Server


class _SessionState:

    def __init__(self, session, hash_funcs):
        """Initialize SessionState instance."""
        self.__dict__["_state"] = {
            "data": {},
            "hash": None,
            "hasher": _CodeHasher(hash_funcs),
            "is_rerun": False,
            "session": session,
        }

    def __call__(self, **kwargs):
        """Initialize state data once."""
        for item, value in kwargs.items():
            if item not in self._state["data"]:
                self._state["data"][item] = value

    def __getitem__(self, item):
        """Return a saved state value, None if item is undefined."""
        return self._state["data"].get(item, None)

    def __getattr__(self, item):
        """Return a saved state value, None if item is undefined."""
        return self._state["data"].get(item, None)

    def __setitem__(self, item, value):
        """Set state value."""
        self._state["data"][item] = value

    def __setattr__(self, item, value):
        """Set state value."""
        self._state["data"][item] = value

    def clear(self):
        """Clear session state and request a rerun."""
        self._state["data"].clear()
        self._state["session"].request_rerun()

    def sync(self):
        """Rerun the app with all state values up to date from the beginning to fix rollbacks."""

        # Ensure to rerun only once to avoid infinite loops
        # caused by a constantly changing state value at each run.
        #
        # Example: state.value += 1
        if self._state["is_rerun"]:
            self._state["is_rerun"] = False

        elif self._state["hash"] is not None:
            if self._state["hash"] != self._state["hasher"].to_bytes(self._state["data"], None):
                self._state["is_rerun"] = True
                self._state["session"].request_rerun()

        self._state["hash"] = self._state["hasher"].to_bytes(
            self._state["data"], None)


def _get_session(session_id=None):
    if not session_id:
        session_id = get_report_ctx().session_id
    session_info = Server.get_current()._get_session_info(session_id)

    if session_info is None:
        raise RuntimeError("Couldn't get your Streamlit Session object.")

    return session_info.session


def get_session_state(session_id=None, hash_funcs=None):
    session = _get_session(session_id)

    if not hasattr(session, "_custom_session_state"):
        session._custom_session_state = _SessionState(session, hash_funcs)

    return session._custom_session_state



from suzieq.sqobjects.basicobj import SqObject
import pandas as pd


class EvpnvniObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='evpnVni', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'vni',
                                'priVtepIp', 'query_str']
        self._valid_assert_args = ['namespace', 'hostname', 'status']
        self._valid_arg_vals = {
            'status': ['all', 'pass', 'fail'],
        }

    def aver(self, **kwargs):
        """Assert that the BGP state is OK"""

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')
        try:
            self.validate_assert_input(**kwargs)
        except Exception as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        return self.engine.aver(**kwargs)



from os.path import isfile, isdir, dirname
import logging
import yaml
import textfsm
from pkgutil import walk_packages
from inspect import getmembers, isclass, getfile
import importlib
from collections import defaultdict
from pathlib import Path

from suzieq.utils import Schema, SchemaForTable
from .service import Service


logger = logging.getLogger(__name__)


def parse_nos_version(svc_dir, filename, svc_def, elem, val):

    if (("command" not in val) or
                ((isinstance(val['command'], list) and not
                  all('textfsm' in x or 'normalize' in x
                      for x in val['command'])) or
                 (not isinstance(val['command'], list) and (
                     "normalize" not in val
                     and "textfsm" not in val)))
            ):
        logger.error(
            "Ignoring invalid service file "
            'definition. Need both "command" and '
            '"normalize/textfsm" keywords: {}, {}'.format(
                filename, val)
        )
        return

    if "textfsm" in val:
        # We may have already visited this element and parsed
        # the textfsm file. Check for this
        if val["textfsm"] and isinstance(
            val["textfsm"], textfsm.TextFSM
        ) or (val["textfsm"] is None):
            return
        tfsm_file = svc_dir + "/" + val["textfsm"]
        if not isfile(tfsm_file):
            logger.error(
                "Textfsm file {} not found. Ignoring"
                " service".format(tfsm_file)
            )
            return
        with open(tfsm_file, "r") as f:
            tfsm_template = textfsm.TextFSM(f)
            val["textfsm"] = tfsm_template
    elif (isinstance(val['command'], list)):
        for subelem in val['command']:
            if 'textfsm' in subelem:
                if subelem["textfsm"] and isinstance(
                    subelem["textfsm"], textfsm.TextFSM
                ):
                    continue
                tfsm_file = svc_dir + "/" + subelem["textfsm"]
                if not isfile(tfsm_file):
                    logger.error(
                        "Textfsm file {} not found. Ignoring"
                        " service".format(tfsm_file)
                    )
                    continue
                with open(tfsm_file, "r") as f:
                    try:
                        tfsm_template = textfsm.TextFSM(f)
                        subelem["textfsm"] = tfsm_template
                    except Exception:
                        logger.exception(
                            'Unable to load TextFSM file '
                            f'{tfsm_file} for service '
                            f'{svc_def["service"]}')
                        continue
    else:
        tfsm_template = None


async def init_services(svc_dir: str, schema_dir: str, queue, svclist: list,
                        def_interval: int, run_once: str):
    """Process service definitions by reading each file in svc dir"""

    svcs_list = []
    schemas = defaultdict(dict)

    # Load up all the service definitions we can find
    svc_classes = {}
    for i in walk_packages(path=[dirname(getfile(Service))]):
        for mbr in getmembers(importlib.import_module(
                'suzieq.poller.services.'+i.name), isclass):
            if mbr[0] == "Service" or not mbr[0].endswith("Service"):
                continue
            svc_classes[i.name] = mbr[1]
            svc_classes[mbr[0]] = mbr[1]

    if not isdir(svc_dir):
        logger.error("services directory not a directory: {}".format(svc_dir))
        return svcs_list

    if not isdir(schema_dir):
        logger.error("schema directory not a directory: {}".format(svc_dir))
        return svcs_list
    else:
        schemas = Schema(schema_dir)

    if schemas:
        poller_schema = schemas.get_arrow_schema("sqPoller")
        poller_schema_version = SchemaForTable('sqPoller', schemas).version

    for filename in Path(svc_dir).glob('*.yml'):
        with open(filename, "r") as f:
            svc_def = yaml.safe_load(f.read())
        if svc_def.get('service') not in svclist:
            logger.warning(
                f'Ignoring unspecified service {svc_def.get("service")}'
            )
            continue

        if "service" not in svc_def or "apply" not in svc_def:
            logger.error(
                'Ignoring invalid service file definition. \
                 "service" and "apply" keywords: {}'.format(
                    filename
                )
            )
            continue

        period = svc_def.get("period", def_interval)
        for elem, val in svc_def["apply"].items():
            if isinstance(val, dict) and "copy" in val:
                newval = svc_def["apply"].get(val["copy"], None)
                if not newval:
                    logger.error(
                        "No device type {} to copy from for "
                        "{} for service {}".format(
                            val["copy"], elem, svc_def["service"]
                        )
                    )
                    return
                val = newval

            if isinstance(val, list):
                for subele in val:
                    parse_nos_version(svc_dir, filename, svc_def, elem,
                                      subele)
            else:

                parse_nos_version(svc_dir, filename, svc_def, elem, val)

        try:
            schema = SchemaForTable(svc_def['service'],
                                    schema=schemas)
        except Exception:
            logger.error(
                f"No matching schema for {svc_def['service']}")
            continue

        if schema.type == "derivedRecord":
            # These are not real services and so ignore them
            continue

        # Valid service definition, add it to list
        if svc_def["service"] in svc_classes:
            service = svc_classes[svc_def["service"]](
                svc_def["service"],
                svc_def["apply"],
                period,
                svc_def.get("type", "state"),
                svc_def.get("keys", []),
                svc_def.get("ignore-fields", []),
                schema,
                queue,
                run_once,
            )
        else:
            service = Service(
                svc_def["service"],
                svc_def["apply"],
                period,
                svc_def.get("type", "state"),
                svc_def.get("keys", []),
                svc_def.get("ignore-fields", []),
                schema,
                queue,
                run_once
            )

        service.poller_schema = poller_schema
        service.poller_schema_version = poller_schema_version
        logger.info("Service {} added".format(service.name))
        svcs_list.append(service)

    return svcs_list

__all__ = [Service, init_services]



from suzieq.sqobjects.basicobj import SqObject


class AddressObj(SqObject):

    def __init__(self, **kwargs):
        super().__init__(table='address', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'address',
                                'columns', 'ipvers', 'vrf', 'query_str']
